"""Additional helper functions for the optlang solvers.

All functions integrate well with the context manager, meaning that
all operations defined here are automatically reverted when used in a
`with model:` block.

The functions defined here together with the existing model functions
should allow you to implement custom flux analysis methods with ease.

"""

import logging
import re
from functools import partial
from types import ModuleType
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Tuple, Union
from warnings import warn

import optlang
import pandas as pd
from optlang.interface import (
    FEASIBLE,
    INFEASIBLE,
    ITERATION_LIMIT,
    NUMERIC,
    OPTIMAL,
    SUBOPTIMAL,
    TIME_LIMIT,
)
from optlang.symbolics import Basic, Zero

from cobra.exceptions import (
    OPTLANG_TO_EXCEPTIONS_DICT,
    OptimizationError,
    SolverNotFound,
)
from cobra.util.context import get_context


# Used to avoid cyclic reference and enable third-party static type checkers to work
if TYPE_CHECKING:
    from cobra import Model, Reaction


CONS_VARS = Union[optlang.interface.Constraint, optlang.interface.Variable]

logger = logging.getLogger(__name__)

# Define all the solvers that are found in optlang.
solvers = {
    match.split("_interface")[0]: getattr(optlang, match)
    for match in dir(optlang)
    if "_interface" in match
}

# Defines all the QP solvers implemented in optlang.
qp_solvers = ["cplex", "gurobi", "osqp"]

# optlang solution statuses which still allow retrieving primal values
has_primals = [NUMERIC, FEASIBLE, INFEASIBLE, SUBOPTIMAL, ITERATION_LIMIT, TIME_LIMIT]


class Components(NamedTuple):
    """Define an object for adding absolute expressions."""

    variable: optlang.interface.Variable
    upper_constraint: optlang.interface.Constraint
    lower_constraint: optlang.interface.Constraint


def linear_reaction_coefficients(
    model: "Model", reactions: Optional[List["Reaction"]] = None
) -> Dict["Reaction", float]:
    """Retrieve coefficient for the reactions in a linear objective.

    Parameters
    ----------
    model : cobra.Model
        The cobra model defining the linear objective.
    reactions : list of cobra.Reaction, optional
        An optional list of the reactions to get the coefficients for.
        By default, all reactions are considered (default None).

    Returns
    -------
    dict
        A dictionary where the keys are the reaction objects and the values
        are the corresponding coefficient. Empty dictionary if there are no
        linear terms in the objective.

    """
    linear_coefficients = {}
    reactions = model.reactions if not reactions else reactions
    try:
        objective_expression = model.solver.objective.expression
        coefficients = objective_expression.as_coefficients_dict()
    except AttributeError:
        return linear_coefficients
    for rxn in reactions:
        forward_coefficient = coefficients.get(rxn.forward_variable, 0)
        reverse_coefficient = coefficients.get(rxn.reverse_variable, 0)
        if forward_coefficient != 0:
            if forward_coefficient == -reverse_coefficient:
                linear_coefficients[rxn] = float(forward_coefficient)
    return linear_coefficients


def _valid_atoms(model: "Model", expression: optlang.symbolics.Basic) -> bool:
    """Check whether a sympy expression references the correct variables.

    Parameters
    ----------
    model : cobra.Model
        The model in which to check for variables.
    expression : sympy.Basic
        A sympy expression.

    Returns
    -------
    bool
        True if all referenced variables are contained in model, False
        otherwise.

    """
    atoms = expression.atoms(optlang.interface.Variable)
    return all(a.problem is model.solver for a in atoms)


def set_objective(
    model: "Model",
    value: Union[
        optlang.interface.Objective,
        optlang.symbolics.Basic,
        Dict["Reaction", float],
    ],
    additive: bool = False,
) -> None:
    """Set the model objective.

    Parameters
    ----------
    model : cobra.Model
       The model to set the objective for.
    value : optlang.interface.Objective, optlang.symbolics.Basic, dict
        If the model objective is linear, then the value can be a new
        optlang.interface.Objective or a dictionary with linear
        coefficients where each key is a reaction and the corresponding
        value is the new coefficient (float).
        If the objective is non-linear and `additive` is True, then only
        values of class optlang.interface.Objective, are accepted.
    additive : bool
        If True, add the terms to the current objective, otherwise start with
        an empty objective.

    Raises
    ------
    ValueError
        If model objective is non-linear and the `value` is a dict.
    TypeError
        If the type of `value` is not one of the accepted ones.

    """
    interface = model.problem
    reverse_value = model.solver.objective.expression
    reverse_value = interface.Objective(
        reverse_value, direction=model.solver.objective.direction, sloppy=True
    )

    if isinstance(value, dict):
        if not model.objective.is_Linear:
            raise ValueError(
                "You can only update non-linear objectives additively using object of "
                f"class optlang.interface.Objective, not of {type(value)}"
            )

        if not additive:
            model.solver.objective = interface.Objective(
                Zero, direction=model.solver.objective.direction
            )
        for reaction, coef in value.items():
            model.solver.objective.set_linear_coefficients(
                {reaction.forward_variable: coef, reaction.reverse_variable: -coef}
            )

    elif isinstance(value, (Basic, optlang.interface.Objective)):
        if isinstance(value, Basic):
            value = interface.Objective(
                value, direction=model.solver.objective.direction, sloppy=False
            )
        # Check whether expression only uses variables from current model;
        # clone the objective if not, faster than cloning without checking
        if not _valid_atoms(model, value.expression):
            value = interface.Objective.clone(value, model=model.solver)

        if not additive:
            model.solver.objective = value
        else:
            model.solver.objective += value.expression
    else:
        raise TypeError(f"{value} is not a valid objective for {model.solver}.")

    context = get_context(model)
    if context:

        def reset():
            model.solver.objective = reverse_value
            model.solver.objective.direction = reverse_value.direction

        context(reset)


def interface_to_str(interface: Union[str, ModuleType]) -> str:
    """Give a string representation for an optlang interface.

    Parameters
    ----------
    interface : str, ModuleType
        Full name of the interface in optlang or cobra representation.
        For instance, 'optlang.glpk_interface' or 'optlang-glpk'.

    Returns
    -------
    str
       The name of the interface as a string.
    """
    if isinstance(interface, ModuleType):
        interface = interface.__name__
    return re.sub(r"optlang.|.interface", "", interface)


def get_solver_name(mip: bool = False, qp: bool = False) -> str:
    """Select a solver for a given optimization problem.

    Parameters
    ----------
    mip : bool
        True if the solver requires mixed integer linear programming capabilities.
    qp : bool
        True if the solver requires quadratic programming capabilities.

    Returns
    -------
    str
        The name of the feasible solver.

    Raises
    ------
    SolverNotFound
        If no suitable solver could be found.

    """
    if len(solvers) == 0:
        raise SolverNotFound("No solvers found.")
    # Those lists need to be updated as optlang implements more solvers
    mip_order = ["gurobi", "cplex", "glpk"]
    lp_order = ["glpk", "cplex", "gurobi"]
    qp_order = ["gurobi", "cplex", "osqp"]

    if mip is False and qp is False:
        for solver_name in lp_order:
            if solver_name in solvers:
                return solver_name
        # none of them are in the list order - so return the first one
        return list(solvers)[0]
    elif qp:  # mip does not yet matter for this determination
        for solver_name in qp_order:
            if solver_name in solvers:
                return solver_name
        raise SolverNotFound("No QP-capable solver found.")
    else:
        for solver_name in mip_order:
            if solver_name in solvers:
                return solver_name
    raise SolverNotFound("No MIP-capable solver found.")


def choose_solver(
    model: "Model", solver: Optional[str] = None, qp: bool = False
) -> ModuleType:
    """Choose a solver given a solver name and model.

    This will choose a solver compatible with the model and required
    capabilities. Also respects model.solver where it can.

    Parameters
    ----------
    model : cobra.Model
        The model for which to choose the solver.
    solver : str, optional
        The name of the solver to be used (default None).
    qp : boolean, optional
        True if the solver needs quadratic programming capabilities
        (default False).

    Returns
    -------
    optlang.interface
        Valid solver for the problem.

    Raises
    ------
    SolverNotFound
        If no suitable solver could be found.

    """
    if solver is None:
        solver = model.problem
    else:
        model.solver = solver

    # Check for QP, raise error if no QP solver found
    if qp and interface_to_str(solver) not in qp_solvers:
        solver = solvers[get_solver_name(qp=True)]

    return solver


def check_solver(obj):
    """Check whether the chosen solver is valid.

    Check whether chosen solver is valid and also warn when using
    a specialized solver. Will return the optlang interface for the
    requested solver.

    Parameters
    ----------
    obj : str or optlang.interface or optlang.interface.Model
        The chosen solver.

    Raises
    ------
    SolverNotFound
        If the solver is not valid.
    """
    not_valid_interface = SolverNotFound(
        f"{obj} is not a valid solver interface. Pick one from {', '.join(solvers)}."
    )
    if isinstance(obj, str):
        try:
            interface = solvers[interface_to_str(obj)]
        except KeyError:
            raise not_valid_interface
    elif isinstance(obj, ModuleType) and hasattr(obj, "Model"):
        interface = obj
    elif isinstance(obj, optlang.interface.Model):
        interface = obj.interface
    else:
        raise not_valid_interface

    if interface_to_str(interface) in ["osqp", "coinor_cbc"]:
        logger.warning(
            "OSQP and CBC are specialized solvers for quadratic programming (QP) and "
            "mixed-integer programming (MIP) problems and may not perform well on "
            "general LP problems. So unless you intend to solve a QP or MIP problem, "
            "we recommend to change the solver back to a general purpose solver "
            "like `model.solver = 'glpk'` for instance."
        )

    return interface


def add_cons_vars_to_problem(
    model: "Model",
    what: Union[List[CONS_VARS], Tuple[CONS_VARS], Components],
    **kwargs,
) -> None:
    """Add variables and constraints to a model's solver object.

    Useful for variables and constraints that can not be expressed with
    reactions and lower/upper bounds. It will integrate with the model's
    context manager in order to revert changes upon leaving the context.

    Parameters
    ----------
    model : cobra.Model
       The model to which to add the variables and constraints.
    what : list or tuple of optlang.interface.Variable or
           optlang.interface.Constraint
       The variables and constraints to add to the model.
    **kwargs : keyword arguments
       Keyword arguments passed to solver's add() method.

    """
    model.solver.add(what, **kwargs)

    context = get_context(model)
    if context:
        context(partial(model.solver.remove, what))


def remove_cons_vars_from_problem(
    model: "Model",
    what: Union[List[CONS_VARS], Tuple[CONS_VARS], Components],
) -> None:
    """Remove variables and constraints from a model's solver object.

    Useful to temporarily remove variables and constraints from a model's
    solver object.

    Parameters
    ----------
    model : cobra.Model
       The model from which to remove the variables and constraints.
    what : list or tuple of optlang.interface.Variable or
           optlang.interface.Constraint
       The variables and constraints to remove from the model.

    """
    model.solver.remove(what)

    context = get_context(model)
    if context:
        context(partial(model.solver.add, what))


def add_absolute_expression(
    model: "Model",
    expression: str,
    name: str = "abs_var",
    ub: Optional[float] = None,
    difference: float = 0.0,
    add: bool = True,
) -> Components:
    """Add the absolute value of an expression to the model.

    Also defines a variable for the absolute value that can be used in
    other objectives or constraints.

    Parameters
    ----------
    model : cobra.Model
       The model to which to add the absolute expression.
    expression : str
       Must be a valid symbolic expression within the model's solver object.
       The absolute value is applied automatically on the expression.
    name : str, optional
       The name of the newly created variable (default "abs_var").
    ub : positive float, optional
       The upper bound for the variable (default None).
    difference : positive float, optional
        The difference between the expression and the variable
        (default 0.0).
    add : bool, optional
        Whether to add the variable to the model at once (default True).

    Returns
    -------
    Components
        A named tuple with variable and two constraints (upper_constraint,
        lower_constraint) describing the new variable and the constraints
        that assign the absolute value of the expression to it.

    """
    variable = model.problem.Variable(name, lb=0, ub=ub)
    # The following constraints enforce variable > expression and
    # variable > -expression
    upper_constraint = model.problem.Constraint(
        expression - variable, ub=difference, name="abs_pos_" + name
    )
    lower_constraint = model.problem.Constraint(
        expression + variable, lb=difference, name="abs_neg_" + name
    )
    to_add = Components(variable, upper_constraint, lower_constraint)
    if add:
        add_cons_vars_to_problem(model, to_add)
    return to_add


def fix_objective_as_constraint(
    model: "Model",
    fraction: float = 1.0,
    bound: Optional[float] = None,
    name: str = "fixed_objective_{}",
) -> float:
    """Fix current objective as an additional constraint.

    When adding constraints to a model, such as done in pFBA which
    minimizes total flux, these constraints can become too powerful,
    resulting in solutions that satisfy optimality but sacrifices too
    much for the original objective function. To avoid that, we can fix
    the current objective value as a constraint to ignore solutions that
    give a lower (or higher depending on the optimization direction)
    objective value than the original model.

    When done with the model as a context, the modification to the
    objective will be reverted when exiting that context.

    Parameters
    ----------
    model : cobra.Model
        The model to operate on.
    fraction : float, optional
        The fraction of the optimum the objective is allowed to reach
        (default 1.0).
    bound : float, optional
        The bound to use instead of fraction of maximum optimal value.
        If not None, `fraction` is ignored (default None).
    name : str, optional
        Name of the objective. May contain one "{}" placeholder which is
        filled with the name of the old objective
        (default "fixed_objective_{}").

    Returns
    -------
    float
        The value of the optimized objective * fraction

    """
    fix_objective_name = name.format(model.objective.name)
    if fix_objective_name in model.constraints:
        model.solver.remove(fix_objective_name)
    if bound is None:
        bound = model.slim_optimize(error_value=None) * fraction
    if model.objective.direction == "max":
        ub, lb = None, bound
    else:
        ub, lb = bound, None
    constraint = model.problem.Constraint(
        model.objective.expression, name=fix_objective_name, ub=ub, lb=lb
    )
    add_cons_vars_to_problem(model, constraint, sloppy=True)
    return bound


def check_solver_status(status: str = None, raise_error: bool = False) -> None:
    """Perform standard checks on a solver's status.

    Parameters
    ----------
    status: str, optional
        The status string obtained from the solver (default None).
    raise_error: bool, optional
        If True, raise error or display warning if False (default False).

    Returns
    -------
    None

    Warns
    -----
    UserWarning
        If `status` is not optimal and `raise_error` is set to True.

    Raises
    ------
    OptimizationError
        If `status` is None or is not optimal and `raise_error` is set to
        True.

    """
    if status == OPTIMAL:
        return None
    elif (status in has_primals) and not raise_error:
        warn(f"Solver status is '{status}'.", UserWarning)
    elif status is None:
        raise OptimizationError(
            "Model is not optimized yet or solver context has been switched."
        )
    else:
        raise OptimizationError(f"Solver status is '{status}'.")


def assert_optimal(model: "Model", message: str = "Optimization failed") -> None:
    """Assert model solver status is optimal.

    Do nothing if model solver status is optimal, otherwise throw
    appropriate exception depending on the status.

    Parameters
    ----------
    model : cobra.Model
        The model to check the solver status for.
    message : str, optional
        Message for the exception if solver status is not optimal
        (default "Optimization failed").

    Returns
    -------
    None

    Raises
    ------
    OptimizationError
       If solver status is not optimal.

    """
    status = model.solver.status
    if status != OPTIMAL:
        exception_cls = OPTLANG_TO_EXCEPTIONS_DICT.get(status, OptimizationError)
        raise exception_cls(f"{message} ({status}).")


def add_lp_feasibility(model: "Model") -> None:
    """Add a new objective and variables to ensure a feasible solution.

    The optimized objective will be zero for a feasible solution and
    otherwise represent the distance from feasibility (please see [1]_
    for more information).

    Parameters
    ----------
    model : cobra.Model
        The model whose feasibility is to be tested.

    Returns
    -------
    None

    References
    ----------
    .. [1] Gomez, Jose A., Kai Höffner, and Paul I. Barton.
    “DFBAlab: A Fast and Reliable MATLAB Code for Dynamic Flux Balance
    Analysis.” BMC Bioinformatics 15, no. 1 (December 18, 2014): 409.
    https://doi.org/10.1186/s12859-014-0409-8.

    """

    obj_vars = []
    prob = model.problem
    for met in model.metabolites:
        s_plus = prob.Variable("s_plus_" + met.id, lb=0)
        s_minus = prob.Variable("s_minus_" + met.id, lb=0)

        model.add_cons_vars([s_plus, s_minus])
        model.constraints[met.id].set_linear_coefficients({s_plus: 1.0, s_minus: -1.0})
        obj_vars.append(s_plus)
        obj_vars.append(s_minus)

    model.objective = prob.Objective(Zero, sloppy=True, direction="min")
    model.objective.set_linear_coefficients({v: 1.0 for v in obj_vars})


def add_lexicographic_constraints(
    model: "Model",
    objectives: List["Reaction"],
    objective_direction: Union[str, List[str]] = "max",
) -> pd.Series:
    """Successively optimize separate targets in a specific order.

    For each objective, optimize the model and set the optimal value as a
    constraint. Proceed in the order of the objectives given. Due to the
    specific order this is called lexicographic FBA [1]_. This procedure
    is useful for returning unique solutions for a set of important
    fluxes. Typically this is applied to exchange fluxes.

    Parameters
    ----------
    model : cobra.Model
        The model to be optimized.
    objectives : list of cobra.Reaction
        A list of reactions (or objectives) in the model for which unique
        fluxes are to be determined.
    objective_direction : str or list of str, optional
        The desired objective direction for each reaction (if a list) or
        the objective direction to use for all reactions (default "max").

    Returns
    -------
    pandas.Series
        A pandas Series containing the optimized fluxes for each of the
        given reactions in `objectives`.

    References
    ----------
    .. [1] Gomez, Jose A., Kai Höffner, and Paul I. Barton.
    “DFBAlab: A Fast and Reliable MATLAB Code for Dynamic Flux Balance
    Analysis.” BMC Bioinformatics 15, no. 1 (December 18, 2014): 409.
    https://doi.org/10.1186/s12859-014-0409-8.

    """

    if type(objective_direction) is not list:
        objective_direction = [objective_direction] * len(objectives)

    constraints = []
    for rxn_id, obj_dir in zip(objectives, objective_direction):
        model.objective = model.reactions.get_by_id(rxn_id)
        model.objective_direction = obj_dir
        constraints.append(fix_objective_as_constraint(model))

    return pd.Series(constraints, index=objectives)



"""Provide unified interfaces to optimization solutions."""

import logging
from typing import TYPE_CHECKING, Iterable, Optional

import numpy as np
import pandas as pd
from optlang.interface import OPTIMAL

from ..util.solver import check_solver_status


if TYPE_CHECKING:
    from cobra import Metabolite, Model, Reaction


__all__ = ("Solution", "get_solution")

logger = logging.getLogger(__name__)


class Solution:
    """
    A unified interface to a `cobra.Model` optimization solution.

    Parameters
    ----------
    objective_value : float
        The (optimal) value for the objective function.
    status : str
        The solver status related to the solution.
    fluxes : pandas.Series
        Contains the reaction fluxes (primal values of variables).
    reduced_costs : pandas.Series
        Contains reaction reduced costs (dual values of variables)
        (default None).
    shadow_prices : pandas.Series
        Contains metabolite shadow prices (dual values of constraints)
        (default None).

    Attributes
    ----------
    objective_value : float
        The (optimal) value for the objective function.
    status : str
        The solver status related to the solution.
    fluxes : pandas.Series
        Contains the reaction fluxes (primal values of variables).
    reduced_costs : pandas.Series
        Contains reaction reduced costs (dual values of variables).
    shadow_prices : pandas.Series
        Contains metabolite shadow prices (dual values of constraints).

    Notes
    -----
    Solution is meant to be constructed by `get_solution` please look at that
    function to fully understand the `Solution` class.

    """

    def __init__(
        self,
        objective_value: float,
        status: str,
        fluxes: pd.Series,
        reduced_costs: Optional[pd.Series] = None,
        shadow_prices: Optional[pd.Series] = None,
        **kwargs,
    ) -> None:
        """
        Initialize a `Solution` from its components.

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        """
        super().__init__(**kwargs)
        self.objective_value = objective_value
        self.status = status
        self.fluxes = fluxes
        self.reduced_costs = reduced_costs
        self.shadow_prices = shadow_prices

    def __repr__(self) -> str:
        """Return a string representation of the solution instance."""
        if self.status != OPTIMAL:
            return f"<Solution {self.status} at {id(self):#x}>"
        return f"<Solution {self.objective_value:.3f} at {id(self):#x}>"

    def _repr_html_(self) -> str:
        """Return a rich HTML representation of the solution."""
        if self.status == OPTIMAL:
            with pd.option_context("display.max_rows", 10):
                html = (
                    "<strong><em>Optimal</em> solution with objective "
                    f"value {self.objective_value:.3f}</strong><br>"
                    f"{self.to_frame()._repr_html_()}"
                )
        else:
            html = f"<strong><em>{self.status}</em> solution</strong>"
        return html

    def __getitem__(self, reaction_id: str) -> float:
        """
        Return the flux of a reaction.

        Parameters
        ----------
        reaction_id : str
            A model reaction ID.

        Returns
        -------
        float
            The flux of the reaction with ID `reaction_id`.

        """
        return self.fluxes[reaction_id]

    get_primal_by_id = __getitem__

    def to_frame(self) -> pd.DataFrame:
        """Return the fluxes and reduced costs as a pandas DataFrame.

        Returns
        -------
        pandas.DataFrame
            The fluxes and reduced cost.

        """
        return pd.DataFrame(
            {"fluxes": self.fluxes, "reduced_costs": self.reduced_costs}
        )


def get_solution(
    model: "Model",
    reactions: Optional[Iterable["Reaction"]] = None,
    metabolites: Optional[Iterable["Metabolite"]] = None,
    raise_error: bool = False,
) -> Solution:
    """
    Generate a solution representation of the current solver state.

    Parameters
    ---------
    model : cobra.Model
        The model whose reactions to retrieve values for.
    reactions : list, optional
        An iterable of `cobra.Reaction` objects. Uses `model.reactions`
        if None (default None).
    metabolites : list, optional
        An iterable of `cobra.Metabolite` objects. Uses `model.metabolites`
        if None (default None).
    raise_error : bool
        If True, raise an OptimizationError if solver status is not optimal
        (default False).

    Returns
    -------
    cobra.Solution

    """
    check_solver_status(model.solver.status, raise_error=raise_error)
    if reactions is None:
        reactions = model.reactions
    if metabolites is None:
        metabolites = model.metabolites

    rxn_index = []
    fluxes = np.empty(len(reactions))
    reduced = np.empty(len(reactions))
    var_primals = model.solver.primal_values
    shadow = np.empty(len(metabolites))
    if model.solver.is_integer:
        reduced.fill(np.nan)
        shadow.fill(np.nan)
        for i, rxn in enumerate(reactions):
            rxn_index.append(rxn.id)
            fluxes[i] = var_primals[rxn.id] - var_primals[rxn.reverse_id]
        met_index = [met.id for met in metabolites]
    else:
        var_duals = model.solver.reduced_costs
        for i, rxn in enumerate(reactions):
            forward = rxn.id
            reverse = rxn.reverse_id
            rxn_index.append(forward)
            fluxes[i] = var_primals[forward] - var_primals[reverse]
            reduced[i] = var_duals[forward] - var_duals[reverse]
        met_index = []
        constr_duals = model.solver.shadow_prices
        for i, met in enumerate(metabolites):
            met_index.append(met.id)
            shadow[i] = constr_duals[met.id]
    return Solution(
        objective_value=model.solver.objective.value,
        status=model.solver.status,
        fluxes=pd.Series(index=rxn_index, data=fluxes, name="fluxes"),
        reduced_costs=pd.Series(index=rxn_index, data=reduced, name="reduced_costs"),
        shadow_prices=pd.Series(index=met_index, data=shadow, name="shadow_prices"),
    )



"""Provide functions for analyzing / creating objective functions."""

from operator import attrgetter
from typing import TYPE_CHECKING, Dict, Union
from warnings import warn

from ..core import Reaction


if TYPE_CHECKING:
    from cobra import Model


def assess(
    model: "Model",
    reaction: Union[Reaction, str],
    flux_coefficient_cutoff: float = 0.001,
    solver: str = None,
) -> Union[bool, Dict]:
    """Assess production capacity.

    Assess the capacity of the `model` to produce the precursors for the
    `reaction` and absorb the production of the `reaction` while the
    `reaction` is operating at, or above, the specified
    `flux_coefficient_cutoff`.

    .. deprecated:: 0.10.0a1
              `solver` argument will be removed in cobrapy 1.0.0, it is
              replaced by globally setting the solver via
              cobra.Configuration .

    Parameters
    ----------
    model : cobra.Model
        The cobra model to assess production capacity for.
    reaction : str or cobra.Reaction
        The reaction to assess.
    flux_coefficient_cutoff : float, optional
        The minimum flux that reaction must carry to be considered active
        (default 0.001).
    solver : str, optional
        The solver name. If None, the default solver will be used
        (default None).

    Returns
    -------
    bool or dict
        True if the model can produce the precursors and absorb the products
        for the reaction operating at, or above, `flux_coefficient_cutoff`.
        Otherwise, a dictionary of {'precursor': Status, 'product': Status},
        where 'Status' is the results from `assess_precursors` and
        `assess_products`, respectively.

    """
    reaction = model.reactions.get_by_any(reaction)[0]
    with model as m:
        m.objective = reaction
        if m.slim_optimize(error_value=0.0) >= flux_coefficient_cutoff:
            return True
        else:
            results = {}
            results["precursors"] = assess_component(
                model, reaction, "reactants", flux_coefficient_cutoff
            )
            results["products"] = assess_component(
                model, reaction, "products", flux_coefficient_cutoff
            )
            return results


def assess_component(
    model: "Model",
    reaction: Union[Reaction, str],
    side: str,
    flux_coefficient_cutoff: float = 0.001,
    solver: str = None,
) -> Union[bool, Dict]:
    """Assess production capacity of components.

    Assess the ability of the `model` to provide sufficient precursors,
    or absorb products, for a `reaction` operating at, or beyond,
    the specified `flux_coefficient_cutoff`.

    .. deprecated:: 0.10.0a1
              `solver` argument will be removed in cobrapy 1.0.0, it is
              replaced by globally setting the solver via
              cobra.Configuration .

    Parameters
    ----------
    model : cobra.Model
        The cobra model to assess production capacity for.
    reaction : str or cobra.Reaction
        The reaction to assess.
    side : {"products", "reactants"}
        The side of the reaction to assess.
    flux_coefficient_cutoff : float, optional
        The minimum flux that reaction must carry to be considered active
        (default 0.001).
    solver : str, optional
        The solver name. If None, the default solver will be used
        (default None).

    Returns
    -------
    bool or dict
        True if the precursors can be simultaneously produced at the
        specified cutoff. False, if the model has the capacity to produce
        each individual precursor at the specified threshold  but not all
        precursors at the required level simultaneously. Otherwise, a
        dictionary of the required and the produced fluxes for each
        reactant that is not produced in sufficient quantities.

    """
    reaction = model.reactions.get_by_any(reaction)[0]
    result_key = {"reactants": "produced", "products": "capacity"}[side]
    get_components = attrgetter(side)
    with model as m:
        m.objective = reaction
        if m.slim_optimize(error_value=0.0) >= flux_coefficient_cutoff:
            return True
        simulation_results = {}
        # build the demand reactions and add all at once
        demand_reactions = {}
        for component in get_components(reaction):
            coeff = reaction.metabolites[component]
            demand = m.add_boundary(component, type="demand")
            demand.metabolites[component] = coeff
            demand_reactions[demand] = (component, coeff)
        # First assess whether all precursors can be produced simultaneously
        joint_demand = Reaction("joint_demand")
        for demand_reaction in demand_reactions:
            joint_demand += demand_reaction
        m.add_reactions([joint_demand])
        m.objective = joint_demand
        if m.slim_optimize(error_value=0.0) >= flux_coefficient_cutoff:
            return True

        # Otherwise assess the ability of the model to produce each precursor
        # individually.  Now assess the ability of the model to produce each
        # reactant for a reaction
        for demand_reaction, (component, coeff) in demand_reactions.items():
            # Calculate the maximum amount of the
            with m:
                m.objective = demand_reaction
                flux = m.slim_optimize(error_value=0.0)
            # metabolite that can be produced.
            if flux_coefficient_cutoff > flux:
                # Scale the results to a single unit
                simulation_results.update(
                    {
                        component: {
                            "required": flux_coefficient_cutoff / abs(coeff),
                            result_key: flux / abs(coeff),
                        }
                    }
                )
        if len(simulation_results) == 0:
            simulation_results = False
        return simulation_results


def _optimize_or_value(model: "Model", value: float = 0.0) -> float:
    """Perform quick optimization and return the objective value.

    The objective value is returned at `value` error value.

    .. deprecated:: 0.22.0
              `_optimize_or_value` will be removed in cobrapy 1.0.0, its
              functionality can be achieved by using
              `cobra.Model.slim_optimize`.

    Parameters
    ----------
    model: cobra.Model
        The model to optimize.
    value: float
        The error value to consider.

    Returns
    -------
    float
        The optimized value of the model's objective.

    """
    return model.slim_optimize(error_value=value)


def assess_precursors(
    model: "Model",
    reaction: Reaction,
    flux_coefficient_cutoff: float = 0.001,
    solver: str = None,
) -> Union[bool, Dict]:
    """Assess production capacity of precursors.

    Assess the ability of the model to provide sufficient precursors for
    a reaction operating at, or beyond, the `flux_coefficient_cutoff`.

    .. deprecated:: 0.7.0
              `assess_precursors` will be removed in cobrapy 1.0.0, it is
              replaced by `assess_component`.

    Parameters
    ----------
    model : cobra.Model
        The cobra model to assess production capacity for.
    reaction : str or cobra.Reaction
        The reaction to assess.
    flux_coefficient_cutoff : float, optional
        The minimum flux that reaction must carry to be considered active
        (default 0.001).
    solver : str, optional
        The solver name. If None, the default solver will be used
        (default None).

    Returns
    -------
    bool or dict
        True if the precursors can be simultaneously produced at the
        specified cutoff. False, if the model has the capacity to produce
        each individual precursor at the specified threshold but not all
        precursors at the required level simultaneously. Otherwise, a
        dictionary of the required and the produced fluxes for each
        reactant that is not produced in sufficient quantities.

    """
    warn("Use cobra.sampling.reaction.assess_component() instead.", DeprecationWarning)
    return assess_component(
        model, reaction, "reactants", flux_coefficient_cutoff, solver
    )


def assess_products(
    model: "Model",
    reaction: Reaction,
    flux_coefficient_cutoff: float = 0.001,
    solver: str = None,
) -> Union[bool, Dict]:
    """Assesses absorption capacity of products.

    Assess the ability of the model to to absorb the products of a reaction
    at a given flux rate. Useful for identifying which components might be
    blocking a reaction from achieving a specific flux rate.

    .. deprecated:: 0.7.0
              `assess_products` will be removed in cobrapy 1.0.0, it is
              replaced by `assess_component`.

    Parameters
    ----------
    model : cobra.Model
        The cobra model to assess production capacity for.
    reaction : str or cobra.Reaction
        The reaction to assess.
    flux_coefficient_cutoff :  float, optional
        The minimum flux that reaction must carry to be considered active
        (default 0.001).
    solver : str, optional
        The solver name. If None, the default solver will be used
        (default None).

    Returns
    -------
    bool or dict
        True if the model has the capacity to absorb all the reaction
        products being simultaneously given the specified cutoff. False,
        if the model has the capacity to absorb each individual product but
        not all products at the required level simultaneously. Otherwise, a
        dictionary of the required and the capacity fluxes for each product
        that is not absorbed in sufficient quantities.

    """
    warn("Use cobra.sampling.reaction.assess_component() instead.", DeprecationWarning)
    return assess_component(
        model, reaction, "products", flux_coefficient_cutoff, solver
    )



"""Provide an implementation of FASTCC."""

from typing import TYPE_CHECKING, List, Optional

from optlang.symbolics import Zero

from .helpers import normalize_cutoff


if TYPE_CHECKING:
    from cobra.core import Model, Reaction


def _find_sparse_mode(
    model: "Model", rxns: List["Reaction"], flux_threshold: float, zero_cutoff: float
) -> List["Reaction"]:
    """Perform the LP required for FASTCC.

    Parameters
    ----------
    model: cobra.Model
        The model to perform FASTCC on.
    rxns: list of cobra.Reaction
        The reactions to use for LP.
    flux_threshold: float
        The upper threshold an auxiliary variable can have.
    zero_cutoff: float
        The cutoff below which flux is considered zero.

    Returns
    -------
    list of cobra.Reaction
        The list of reactions to consider as consistent.

    """
    if rxns:
        obj_vars = []
        vars_and_cons = []
        prob = model.problem

        for rxn in rxns:
            var = prob.Variable(
                "auxiliary_{}".format(rxn.id), lb=0.0, ub=flux_threshold
            )
            const = prob.Constraint(
                rxn.forward_variable + rxn.reverse_variable - var,
                name="constraint_{}".format(rxn.id),
                lb=0.0,
            )
            vars_and_cons.extend([var, const])
            obj_vars.append(var)

        model.add_cons_vars(vars_and_cons)
        model.objective = prob.Objective(Zero, sloppy=True)
        model.objective.set_linear_coefficients({v: 1.0 for v in obj_vars})

        model.optimize(objective_sense="max")
        result = [rxn for rxn in model.reactions if abs(rxn.flux) > zero_cutoff]
    else:
        result = []

    return result


def _flip_coefficients(model: "Model", rxns: List["Reaction"]) -> None:
    """Flip the coefficients for optimizing in reverse direction.

    Parameters
    ----------
    model: cobra.Model
        The model to operate on.
    rxns: list of cobra.Reaction
        The list of reactions whose coefficients will be flipped.

    """
    # flip reactions
    for rxn in rxns:
        const = model.constraints.get("constraint_{}".format(rxn.id))
        var = model.variables.get("auxiliary_{}".format(rxn.id))
        coefs = const.get_linear_coefficients(const.variables)
        const.set_linear_coefficients({k: -v for k, v in coefs.items() if k is not var})

    # flip objective
    objective = model.objective
    objective_coefs = objective.get_linear_coefficients(objective.variables)
    objective.set_linear_coefficients({k: -v for k, v in objective_coefs.items()})


def fastcc(
    model: "Model", flux_threshold: float = 1.0, zero_cutoff: Optional[float] = None
) -> "Model":
    r"""
    Check consistency of a metabolic network using FASTCC [1]_.

    FASTCC (Fast Consistency Check) is an algorithm for rapid and
    efficient consistency check in metabolic networks. FASTCC is
    a pure LP implementation and is low on computation resource
    demand. FASTCC also circumvents the problem associated with
    reversible reactions for the purpose. Given a global model,
    it will generate a consistent global model i.e., remove
    blocked reactions. For more details on FASTCC, please
    check [1]_.

    Parameters
    ----------
    model: cobra.Model
        The model to operate on.
    flux_threshold: float, optional
        The flux threshold to consider (default 1.0).
    zero_cutoff: float, optional
        The cutoff to consider for zero flux (default model.tolerance).

    Returns
    -------
    cobra.Model
        The consistent model.

    Notes
    -----
    The LP used for FASTCC is like so:
    maximize: \sum_{i \in J} z_i
    s.t.    : z_i \in [0, \varepsilon] \forall i \in J, z_i \in \mathbb{R}_+
              v_i \ge z_i \forall i \in J
              Sv = 0 v \in B

    References
    ----------
    .. [1] Vlassis N, Pacheco MP, Sauter T (2014)
           Fast Reconstruction of Compact Context-Specific Metabolic Network
           Models.
           PLoS Comput Biol 10(1): e1003424. doi:10.1371/journal.pcbi.1003424

    """
    zero_cutoff = normalize_cutoff(model, zero_cutoff)

    irreversible_rxns = [rxn for rxn in model.reactions if not rxn.reversibility]
    rxns_to_check = irreversible_rxns

    with model:
        rxns_to_keep = _find_sparse_mode(
            model, rxns_to_check, flux_threshold, zero_cutoff
        )

    rxns_to_check = list(set(model.reactions).difference(rxns_to_keep))

    while rxns_to_check:
        with model:
            new_rxns = _find_sparse_mode(
                model, rxns_to_check, flux_threshold, zero_cutoff
            )
            rxns_to_keep.extend(new_rxns)

            # this condition will be valid for all but the last iteration
            if list(set(rxns_to_check).intersection(rxns_to_keep)):
                rxns_to_check = list(set(rxns_to_check).difference(rxns_to_keep))

            else:
                rxns_to_flip = list(set(rxns_to_check).difference(irreversible_rxns))
                _flip_coefficients(model, rxns_to_flip)
                sol = model.optimize(min)
                to_add_rxns = sol.fluxes.index[sol.fluxes.abs() > zero_cutoff].tolist()
                rxns_to_keep.extend(
                    [model.reactions.get_by_id(rxn) for rxn in to_add_rxns]
                )
                # since this is the last iteration, it needs to break or else
                # it will run forever since rxns_to_check won't be empty
                break

    consistent_rxns = set(rxns_to_keep)
    # need the ids since Reaction objects are created fresh with model.copy()
    rxns_to_remove = [
        rxn.id for rxn in set(model.reactions).difference(consistent_rxns)
    ]

    consistent_model = model.copy()
    consistent_model.remove_reactions(rxns_to_remove, remove_orphans=True)

    return consistent_model



"""
SBML import and export using python-libsbml.

- The SBML importer supports all versions of SBML and the fbc package.
- The SBML exporter writes SBML L3 models.
- Annotation information is stored on the cobrapy objects
- Information from the group package is read

Parsing of fbc models was implemented as efficient as possible, whereas
(discouraged) fallback solutions are not optimized for efficiency.

Notes are only supported in a minimal way relevant for constraint-based
models. I.e., structured information from notes in the form
   <p>key: value</p>
is read into the Object.notes dictionary when reading SBML files.
On writing the Object.notes dictionary is serialized to the SBML
notes information.

Annotations are read in the Object.annotation fields.

Some SBML related issues are still open, please refer to the respective issue:
- update annotation format and support qualifiers (depends on decision
    for new annotation format; https://github.com/opencobra/cobrapy/issues/684)
- write compartment annotations and notes (depends on updated first-class
    compartments; see https://github.com/opencobra/cobrapy/issues/760)
- support compression on file handles (depends on solution for
    https://github.com/opencobra/cobrapy/issues/812)
"""

import datetime
import logging
import re
from ast import And, BoolOp, Module, Name, Or
from collections import defaultdict, namedtuple
from copy import deepcopy
from io import StringIO
from pathlib import Path
from sys import platform
from typing import IO, Match, Optional, Pattern, Tuple, Type, Union

import libsbml

import cobra

from ..core import GPR, Gene, Group, Metabolite, Model, Reaction
from ..manipulation.validate import check_metabolite_compartment_formula
from ..util.solver import linear_reaction_coefficients, set_objective


class CobraSBMLError(Exception):
    """SBML error class."""

    pass


LOGGER = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Defaults and constants for writing SBML
# -----------------------------------------------------------------------------
config = cobra.Configuration()  # for default bounds
LOWER_BOUND_ID = "cobra_default_lb"
UPPER_BOUND_ID = "cobra_default_ub"
ZERO_BOUND_ID = "cobra_0_bound"

BOUND_MINUS_INF = "minus_inf"
BOUND_PLUS_INF = "plus_inf"

SBO_FBA_FRAMEWORK = "SBO:0000624"
SBO_DEFAULT_FLUX_BOUND = "SBO:0000626"
SBO_FLUX_BOUND = "SBO:0000625"
SBO_EXCHANGE_REACTION = "SBO:0000627"

LONG_SHORT_DIRECTION = {"maximize": "max", "minimize": "min"}
SHORT_LONG_DIRECTION = {"min": "minimize", "max": "maximize"}

Unit = namedtuple("Unit", ["kind", "scale", "multiplier", "exponent"])
UNITS_FLUX = (
    "mmol_per_gDW_per_hr",
    [
        Unit(kind=libsbml.UNIT_KIND_MOLE, scale=-3, multiplier=1, exponent=1),
        Unit(kind=libsbml.UNIT_KIND_GRAM, scale=0, multiplier=1, exponent=-1),
        Unit(kind=libsbml.UNIT_KIND_SECOND, scale=0, multiplier=3600, exponent=-1),
    ],
)

# -----------------------------------------------------------------------------
# Functions for id replacements (import/export)
# -----------------------------------------------------------------------------
SBML_DOT = "__SBML_DOT__"

# -----------------------------------------------------------------------------
# Precompiled note pattern
# -----------------------------------------------------------------------------
pattern_notes: Pattern = re.compile(
    r"<(?P<prefix>(\w+:)?)p[^>]*>(?P<content>.*?)</(?P=prefix)p>",
    re.IGNORECASE | re.DOTALL,
)

pattern_to_sbml: Pattern = re.compile(r"([^0-9_a-zA-Z])")

pattern_from_sbml: Pattern = re.compile(r"__(\d+)__")


def _escape_non_alphanum(nonASCII: Match) -> str:
    """Convert non alphanumeric to string representation of ascii number.

    Converts a non alphanumeric character to a string representation of
    its ascii number using ord.

    Parameters
    ---------
    nonASCII: Match
        Match object, identified by pattern_from_sbml

    Returns
    -------
    str:
        The ascii code, surronded by __.
    """
    return "__" + str(ord(nonASCII.group())) + "__"


def _number_to_chr(numberStr: Match) -> str:
    """Convert an ascii number to a character.

    Parameters
    ---------
    numberStr: Match
        Match object, identified by pattern_from_sbml

    Returns
    -------
    str:
        The first match between the underscores, converted to character.
    """
    return chr(int(numberStr.group(1)))


def _clip(sid: str, prefix: str) -> str:
    """Clip a prefix from the beginning of a string if it exists.

    Parameters
    ----------
    sid: str
        String to clip.
    prefix: str
        Prefix to remove.

    Returns
    -------
    str
        The string with prefix clipped if it existed. Otherwise the original string.
    """
    return sid[len(prefix) :] if sid.startswith(prefix) else sid


def _f_gene(sid: str, prefix: str = "G_") -> str:
    """Clip gene prefix from id.

    Parameters
    ----------
    sid: str
        String to process
    prefix: str, optional
        Prefix for gene (default "G_").

    Returns
    -------
    str
        Returns modified str with prefix removed, SBML_DOT (see above) replaced
        with ".", __(NUMBER)__ replaced with the character value of NUMBER.

    See Also
    --------
    pattern_from_sbml
    _number_to_chr
    SBML_DOT
    """
    sid = sid.replace(SBML_DOT, ".")
    sid = pattern_from_sbml.sub(_number_to_chr, sid)
    return _clip(sid, prefix)


def _f_gene_rev(sid: str, prefix: str = "G_") -> str:
    """Add gene prefix to id.

    Parameters
    ----------
    sid: str
        String to process
    prefix: str, optional
        Prefix for gene (default "G_").

    Returns
    -------
    str
        Returns prefix prepended to modified str, with "." replaced with
        SBML_DOT, non alphanumeric repalced with a string representation of the
        unicode number.

    See Also
    --------
    pattern_to_sbml
    _escape_non_alphanum
    SBML_DOT
    """
    sid = pattern_to_sbml.sub(_escape_non_alphanum, sid)
    return prefix + sid.replace(".", SBML_DOT)


def _f_specie(sid: str, prefix: str = "M_") -> str:
    """Clip specie/metabolite prefix from id.

    Parameters
    ----------
    sid: str
        String to process
    prefix: str, optional
        Prefix for species/metabolite (default "M_").

    Returns
    -------
    str
        Returns modified str with prefix removed, SBML_DOT (see above) replaced
        with ".", __(NUMBER)__ replaced with the character value of NUMBER.

    See Also
    --------
    pattern_from_sbml
    _number_to_chr
    SBML_DOT
    """
    sid = pattern_from_sbml.sub(_number_to_chr, sid)
    return _clip(sid, prefix)


def _f_specie_rev(sid: str, prefix: str = "M_") -> str:
    """Add specie/metabolite prefix to id.

    Parameters
    ----------
    sid: str
        String to process
    prefix: str, optional
        Prefix for metabolite (default "M_").

    Returns
    -------
    str
        Returns prefix prepended to modified str, with "." replaced with
        SBML_DOT, non alphanumeric repalced with a string representation of the
        unicode number.

    See Also
    --------
    pattern_to_sbml
    _escape_non_alphanum
    SBML_DOT
    """
    sid = pattern_to_sbml.sub(_escape_non_alphanum, sid)
    return prefix + sid


def _f_reaction(sid: str, prefix: str = "R_") -> str:
    """Clip reaction prefix from id.

    Parameters
    ----------
    sid: str
        String to process
    prefix: str, optional
        Prefix for reaction (default "R_").

    Returns
    -------
    str
        Returns modified str with prefix removed, SBML_DOT (see above) replaced
        with ".", __(NUMBER)__ replaced with the character value of NUMBER.

    See Also
    --------
    pattern_from_sbml
    _number_to_chr
    SBML_DOT
    """
    sid = pattern_from_sbml.sub(_number_to_chr, sid)
    return _clip(sid, prefix)


def _f_reaction_rev(sid: str, prefix: str = "R_") -> str:
    """Add reaction prefix to id.

    Parameters
    ----------
    sid: str
        String to process
    prefix: str, optional
        Prefix for reaction (default "R_").

    Returns
    -------
    str
        Returns prefix prepended to modified str, with "." replaced with
        SBML_DOT, non alphanumeric repalced with a string representation of the
        unicode number.

    See Also
    --------
    pattern_to_sbml
    _escape_non_alphanum
    SBML_DOT
    """
    sid = pattern_to_sbml.sub(_escape_non_alphanum, sid)
    return prefix + sid


def _f_group(sid: str, prefix: str = "G_") -> str:
    """Clip group prefix from id.

    Parameters
    ----------
    sid: str
        String to process
    prefix: str, optional
        Prefix for group (default "G_").

    Returns
    -------
    str
        Returns modified str with prefix removed, SBML_DOT (see above) replaced
        with ".", __(NUMBER)__ replaced with the character value of NUMBER.

    See Also
    --------
    pattern_from_sbml
    _number_to_chr
    SBML_DOT
    """
    sid = pattern_from_sbml.sub(_number_to_chr, sid)
    return _clip(sid, prefix)


def _f_group_rev(sid: str, prefix: str = "G_") -> str:
    """Add group prefix to id.

    Parameters
    ----------
    sid: str
        String to process
    prefix: str, optional
        Prefix for group (default "G_").

    Returns
    -------
    str
        Returns prefix prepended to modified str, with "." replaced with
        SBML_DOT, non alphanumeric repalced with a string representation of the
        unicode number.

    See Also
    --------
    pattern_to_sbml
    _escape_non_alphanum
    SBML_DOT
    """
    sid = pattern_to_sbml.sub(_escape_non_alphanum, sid)
    return prefix + sid


F_GENE = "F_GENE"
F_GENE_REV = "F_GENE_REV"
F_SPECIE = "F_SPECIE"
F_SPECIE_REV = "F_SPECIE_REV"
F_REACTION = "F_REACTION"
F_REACTION_REV = "F_REACTION_REV"
F_GROUP = "F_GROUP"
F_GROUP_REV = "F_GROUP_REV"

F_REPLACE: dict = {
    F_GENE: _f_gene,
    F_GENE_REV: _f_gene_rev,
    F_SPECIE: _f_specie,
    F_SPECIE_REV: _f_specie_rev,
    F_REACTION: _f_reaction,
    F_REACTION_REV: _f_reaction_rev,
    F_GROUP: _f_group,
    F_GROUP_REV: _f_group_rev,
}


# -----------------------------------------------------------------------------
# Read SBML
# -----------------------------------------------------------------------------
# noinspection PyDefaultArgument
def read_sbml_model(
    filename: Union[str, IO, Path],
    number: Type = float,
    f_replace: dict = F_REPLACE,
    **kwargs,
) -> Model:
    """Read SBML model from given filename.

    If the given filename ends with the suffix ''.gz'' (for example,
    ''myfile.xml.gz'),' the file is assumed to be compressed in gzip
    format and will be automatically decompressed upon reading. Similarly,
    if the given filename ends with ''.zip'' or ''.bz2',' the file is
    assumed to be compressed in zip or bzip2 format (respectively).  Files
    whose names lack these suffixes will be read uncompressed.  Note that
    if the file is in zip format but the archive contains more than one
    file, only the first file in the archive will be read and the rest
    ignored.

    To read a gzip/zip file, libSBML needs to be configured and linked
    with the zlib library at compile time.  It also needs to be linked
    with the bzip2 library to read files in bzip2 format.  (Both of these
    are the default configurations for libSBML.)

    This function supports SBML with FBC-v1 and FBC-v2. FBC-v1 models
    are converted to FBC-v2 models before reading.

    The parser tries to fall back to information in notes dictionaries
    if information is not available in the FBC packages, e.g.,
    CHARGE, FORMULA on species, or GENE_ASSOCIATION, SUBSYSTEM on reactions.

    Parameters
    ----------
    filename : path to SBML file, or SBML string, or SBML file handle
        SBML which is read into cobra model
    number: data type of stoichiometry: {float, int}
        In which data type should the stoichiometry be parsed.
    f_replace : dict of replacement functions for id replacement
        Dictionary of replacement functions for gene, specie, and reaction.
        By default the following id changes are performed on import:
        clip G_ from genes, clip M_ from species, clip R_ from reactions
        If no replacements should be performed, set f_replace={}, None
    **kwargs:
        Further keyword arguments are passed on to the called function (_sbml_to_model)

    Returns
    -------
    cobra.core.Model

    Raises
    ------
    IOError if file not read
    All other errors are wrapped around in a message pointing to SBML validator.

    Notes
    -----
    Provided file handles cannot be opened in binary mode, i.e., use
        with open(path, "r" as f):
            read_sbml_model(f)
    File handles to compressed files are not supported yet.
    """
    try:
        doc = _get_doc_from_filename(filename)
        return _sbml_to_model(doc, number=number, f_replace=f_replace, **kwargs)
    except IOError as e:
        raise e

    except Exception as original_error:
        cobra_error = CobraSBMLError(
            "Something went wrong reading the SBML model. Most likely the SBML"
            " model is not valid. Please check that your model is valid using "
            "the `cobra.io.sbml.validate_sbml_model` function or via the "
            "online validator at https://sbml.org/validator_servlet/ .\n"
            "\t`(model, errors) = validate_sbml_model(filename)`"
            "\nIf the model is valid and cannot be read please open an issue "
            "at https://github.com/opencobra/cobrapy/issues ."
        )
        raise cobra_error from original_error


def _get_doc_from_filename(filename: Union[str, IO, Path]) -> "libsbml.SBMLDocument":
    """Get SBMLDocument from given filename.

    Parameters
    ----------
    filename : path to SBML, or SBML string, or filehandle

    Returns
    -------
    libsbml.SBMLDocument

    Raises
    ------
    IOError if file not readable or does not contain SBML.
    CobraSBMLError if input type is not valid.
    """
    if isinstance(filename, Path) and {".bz2", ".gz"}.isdisjoint(filename.suffixes):
        doc: libsbml.SBMLDocument = libsbml.readSBMLFromString(filename.read_text())
    elif isinstance(filename, Path) and {".bz2", ".gz"}.intersection(filename.suffixes):
        doc: libsbml.SBMLDocument = libsbml.readSBMLFromFile(str(filename))
    elif isinstance(filename, str):
        if "<sbml" in filename:
            doc: libsbml.SBMLDocument = libsbml.readSBMLFromString(filename)
        elif (
            ("win" in platform) and (len(filename) < 260) or "win" not in platform
        ) and Path(filename).exists():
            doc: libsbml.SBMLDocument = libsbml.readSBMLFromFile(filename)
        else:
            # string representation
            raise IOError(
                f"The file with '{filename}' does not exist, "
                f"or is not an SBML string. Provide the path to "
                f"an existing SBML file or a valid SBML string representation:\n"
            )
    elif hasattr(filename, "read"):
        # file handle
        doc: libsbml.SBMLDocument = libsbml.readSBMLFromString(filename.read())
    else:
        raise CobraSBMLError(
            f"Input type '{type(filename)}' for '{filename}' is not supported."
            f" Provide a path, SBML str, or file handle."
        )

    return doc


# noinspection PyDefaultArgument
# noinspection PyUnusedLocal
def _sbml_to_model(
    doc: "libsbml.SBMLDocument",
    number: Type = float,
    f_replace: dict = F_REPLACE,
    set_missing_bounds: bool = False,
    **kwargs,
) -> Model:
    """Create cobra model from SBMLDocument.

    Parameters
    ----------
    doc: libsbml.SBMLDocument
    number: data type of stoichiometry: {float, int}
        In which data type should the stoichiometry be parsed.
    f_replace : dict
        dict of replacement functions for id replacement
    set_missing_bounds : bool
        flag to set missing bounds. Looks like it will be ignored.
    **kwargs:
        Further keyword arguments are passed on.

    Returns
    -------
    cobra.core.Model

    Raises
    ------
    CobraSBMLError if no SBML model detected in file.
    Exception if fbc coversion from v1 to v2 needed and not successful.
    CobraSBMLError if upper or lower bound are missing from a reaction.
    CobraSBMLError if objective reaction declared and not found. Objective reaction not
                    declared will lead to an ERROR being logged.

    """
    if f_replace is None:
        f_replace = {}

    # SBML model
    model: "libsbml.Model" = doc.getModel()
    if model is None:
        raise CobraSBMLError("No SBML model detected in file.")
    model_fbc: "libsbml.FbcModelPlugin" = model.getPlugin("fbc")

    if not model_fbc:
        LOGGER.warning("Model does not contain SBML fbc package information.")
    else:
        if not model_fbc.isSetStrict():
            LOGGER.warning('Loading SBML model without fbc:strict="true"')

        # fbc-v1 (legacy)
        doc_fbc: "libsbml.FbcSBMLDocumentPlugin" = doc.getPlugin("fbc")
        fbc_version = doc_fbc.getPackageVersion()

        if fbc_version == 1:
            LOGGER.warning(
                "Loading SBML with fbc-v1 (models should be encoded using fbc-v2)"
            )
            conversion_properties = libsbml.ConversionProperties()
            conversion_properties.addOption(
                "convert fbc v1 to fbc v2", True, "Convert FBC-v1 model to FBC-v2"
            )
            result = doc.convert(conversion_properties)
            if result != libsbml.LIBSBML_OPERATION_SUCCESS:
                raise Exception("Conversion of SBML fbc v1 to fbc v2 failed")

    # Model
    model_id = model.getIdAttribute()
    if not libsbml.SyntaxChecker.isValidSBMLSId(model_id):
        LOGGER.error(f"'{model_id}' is not a valid SBML 'SId'.")
    cobra_model = Model(model_id)
    cobra_model.name = model.getName() or None

    # meta information
    meta = {
        "model.id": model_id,
        "level": model.getLevel(),
        "version": model.getVersion(),
        "packages": [],
    }
    # History
    creators = []
    created = None
    if model.isSetModelHistory():
        history: "libsbml.ModelHistory" = model.getModelHistory()

        if history.isSetCreatedDate():
            created = history.getCreatedDate().getDateAsString()

        c: "libsbml.ModelCreator"
        for c in history.getListCreators():
            creators.append(
                {
                    "familyName": c.getFamilyName() if c.isSetFamilyName() else None,
                    "givenName": c.getGivenName() if c.isSetGivenName() else None,
                    "organisation": c.getOrganisation()
                    if c.isSetOrganisation()
                    else None,
                    "email": c.getEmail() if c.isSetEmail() else None,
                }
            )

    meta["creators"] = creators
    meta["created"] = created
    meta["notes"] = _parse_notes_dict(doc)
    meta["annotation"] = _parse_annotations(doc)

    info = f"<{model_id}> SBML L{model.getLevel()}V{model.getVersion()}"
    packages = {}
    for k in range(doc.getNumPlugins()):
        plugin: "libsbml.SBasePlugin" = doc.getPlugin(k)
        key, value = plugin.getPackageName(), plugin.getPackageVersion()
        packages[key] = value
        info += f", {key}-v{value}"
        if key not in ["fbc", "groups", "l3v2extendedmath"]:
            LOGGER.warning(
                f"SBML package '{key}' not supported by cobrapy, "
                f"information is not parsed"
            )
    meta["info"] = info
    meta["packages"] = packages
    cobra_model._sbml = meta

    # notes and annotations
    cobra_model.notes = _parse_notes_dict(model)
    cobra_model.annotation = _parse_annotations(model)

    # Compartments
    # FIXME: update with new compartments
    compartments = {}
    for (
        compartment
    ) in model.getListOfCompartments():  # noqa: E501 type: libsbml.Compartment
        cid = _check_required(compartment, compartment.getIdAttribute(), "id")
        compartments[cid] = compartment.getName()
    cobra_model.compartments = compartments

    # Species
    metabolites = []
    boundary_metabolites = []
    if model.getNumSpecies() == 0:
        LOGGER.warning("No metabolites in model")

    specie: "libsbml.Species"
    for specie in model.getListOfSpecies():
        sid = _check_required(specie, specie.getIdAttribute(), "id")
        if f_replace and F_SPECIE in f_replace:
            sid = f_replace[F_SPECIE](sid)

        met = Metabolite(sid)
        met.name = specie.getName()
        met.notes = _parse_notes_dict(specie)
        met.annotation = _parse_annotations(specie)
        met.compartment = specie.getCompartment()

        specie_fbc: "libsbml.FbcSpeciesPlugin" = specie.getPlugin("fbc")
        if specie_fbc:
            met.charge = specie_fbc.getCharge()
            met.formula = specie_fbc.getChemicalFormula() or None
        else:
            if specie.isSetCharge():
                LOGGER.warning(
                    f"Use of the species charge attribute is "
                    f"discouraged, use fbc:charge instead: {specie}"
                )
                met.charge = specie.getCharge()
            else:
                if "CHARGE" in met.notes:
                    LOGGER.warning(
                        f"Use of CHARGE in the notes element is "
                        f"discouraged, use fbc:charge instead: {specie}"
                    )
                    try:
                        met.charge = int(met.notes["CHARGE"])
                    except ValueError:
                        # handle nan, na, NA, ...
                        pass

            if "FORMULA" in met.notes:
                LOGGER.warning(
                    f"Use of FORMULA in the notes element is "
                    f"discouraged, use fbc:chemicalFormula instead: {specie}"
                )
                met.formula = met.notes["FORMULA"] or None

        # Detect boundary metabolites
        if specie.getBoundaryCondition() is True:
            boundary_metabolites.append(met)

        metabolites.append(met)

    cobra_model.add_metabolites(metabolites)

    # Add exchange reactions for boundary metabolites
    ex_reactions = []
    for met in boundary_metabolites:
        ex_rid = f"EX_{met.id}"
        ex_reaction = Reaction(ex_rid)
        ex_reaction.name = ex_rid
        ex_reaction.annotation = {"sbo": SBO_EXCHANGE_REACTION}
        ex_reaction.lower_bound = config.lower_bound
        ex_reaction.upper_bound = config.upper_bound
        LOGGER.warning(
            f"Adding exchange reaction {ex_reaction.id} with default bounds for "
            f"boundary metabolite: {met.id}."
        )
        # species is reactant
        ex_reaction.add_metabolites({met: -1})
        ex_reactions.append(ex_reaction)
    cobra_model.add_reactions(ex_reactions)

    # Genes
    if model_fbc:
        for (
            gp
        ) in model_fbc.getListOfGeneProducts():  # noqa: E501 type: libsbml.GeneProduct
            gid = _check_required(gp, gp.getIdAttribute(), "id")
            if f_replace and F_GENE in f_replace:
                gid = f_replace[F_GENE](gid)
            cobra_gene = Gene(gid)
            cobra_gene.name = gp.getName()
            if cobra_gene.name is None:
                cobra_gene.name = gid
            cobra_gene.annotation = _parse_annotations(gp)
            cobra_gene.notes = _parse_notes_dict(gp)

            cobra_model.genes.append(cobra_gene)
    else:
        for (
            cobra_reaction
        ) in model.getListOfReactions():  # noqa: E501 type: libsbml.Reaction
            # fallback to notes information
            notes = _parse_notes_dict(cobra_reaction)
            if "GENE ASSOCIATION" in notes:
                gpr = notes["GENE ASSOCIATION"]
            elif "GENE_ASSOCIATION" in notes:
                gpr = notes["GENE_ASSOCIATION"]
            else:
                gpr = ""

            if len(gpr) > 0:
                gpr = gpr.replace("(", ";")
                gpr = gpr.replace(")", ";")
                gpr = gpr.replace("or", ";")
                gpr = gpr.replace("and", ";")
                # Interaction of the above replacements can lead to multiple
                # ;, which results in empty gids
                gids = [t.strip() for t in gpr.split(";")]
                gids = set(gids).difference({""})

                # create missing genes
                for gid in gids:
                    if f_replace and F_GENE in f_replace:
                        gid = f_replace[F_GENE](gid)

                    if gid not in cobra_model.genes:
                        cobra_gene = Gene(gid)
                        cobra_gene.name = gid
                        cobra_model.genes.append(cobra_gene)

    # GPR rules
    def process_association(ass: "libsbml.FbcAssociation") -> Union[BoolOp, Name]:
        """Recursively convert gpr association to a GPR class.

        Defined as inline functions to not pass the replacement dict around.

        Parameters
        ----------
        ass: libsbml.FbcAssociation

        Returns
        -------
        BoolOp or Name
            AST formatted of the FbcAssociation, which will be processed by GPR().
        """
        if ass.isFbcOr():
            return BoolOp(
                Or(), [process_association(c) for c in ass.getListOfAssociations()]
            )
        elif ass.isFbcAnd():
            return BoolOp(
                And(), [process_association(c) for c in ass.getListOfAssociations()]
            )
        elif ass.isGeneProductRef():
            g_id = ass.getGeneProduct()
            if f_replace and F_GENE in f_replace:
                return Name(id=f_replace[F_GENE](g_id))
            else:
                return Name(id=g_id)

    # Reactions
    missing_bounds = False
    reactions = []
    if model.getNumReactions() == 0:
        LOGGER.warning("No reactions in model")

    reaction: "libsbml.Reaction"
    for reaction in model.getListOfReactions():
        rid = _check_required(reaction, reaction.getIdAttribute(), "id")
        if f_replace and F_REACTION in f_replace:
            rid = f_replace[F_REACTION](rid)
        cobra_reaction = Reaction(rid)
        cobra_reaction.name = reaction.getName().strip()
        cobra_reaction.annotation = _parse_annotations(reaction)
        cobra_reaction.notes = _parse_notes_dict(reaction)

        # set bounds
        p_ub, p_lb = None, None
        r_fbc: "libsbml.FbcReactionPlugin" = reaction.getPlugin("fbc")
        if r_fbc:
            # bounds in fbc
            lb_id = r_fbc.getLowerFluxBound()
            if lb_id:
                p_lb: "libsbml.Parameter" = model.getParameter(lb_id)
                if p_lb and p_lb.getConstant() and (p_lb.getValue() is not None):
                    cobra_reaction.lower_bound = p_lb.getValue()
                else:
                    raise CobraSBMLError(
                        f"No constant bound '{p_lb}' for reaction: {reaction}"
                    )

            ub_id = r_fbc.getUpperFluxBound()
            if ub_id:
                p_ub: "libsbml.Parameter" = model.getParameter(ub_id)
                if p_ub and p_ub.getConstant() and (p_ub.getValue() is not None):
                    cobra_reaction.upper_bound = p_ub.getValue()
                else:
                    raise CobraSBMLError(
                        f"No constant bound '{p_ub}' for reaction: {reaction}"
                    )

        elif reaction.isSetKineticLaw():
            # some legacy models encode bounds in kinetic laws
            klaw: "libsbml.KineticLaw" = reaction.getKineticLaw()
            p_lb = klaw.getParameter(
                "LOWER_BOUND"
            )  # noqa: E501 type: libsbml.LocalParameter
            if p_lb:
                cobra_reaction.lower_bound = p_lb.getValue()
            p_ub = klaw.getParameter(
                "UPPER_BOUND"
            )  # noqa: E501 type: libsbml.LocalParameter
            if p_ub:
                cobra_reaction.upper_bound = p_ub.getValue()

            if p_ub is not None or p_lb is not None:
                LOGGER.warning(
                    f"Encoding LOWER_BOUND and UPPER_BOUND in "
                    f"KineticLaw is discouraged, "
                    f"use fbc:fluxBounds instead: {reaction}"
                )

        if p_lb is None:
            missing_bounds = True
            lower_bound = config.lower_bound
            cobra_reaction.lower_bound = lower_bound
            LOGGER.warning(
                f"Missing lower flux bound set to '{lower_bound}' for "
                f"reaction: '{reaction}'"
            )

        if p_ub is None:
            missing_bounds = True
            upper_bound = config.upper_bound
            cobra_reaction.upper_bound = upper_bound
            LOGGER.warning(
                f"Missing upper flux bound set to '{upper_bound}' for "
                f"reaction: '{reaction}'"
            )

        # add reaction
        reactions.append(cobra_reaction)

        # parse equation
        stoichiometry = defaultdict(lambda: 0)
        for (
            sref
        ) in reaction.getListOfReactants():  # noqa: E501 type: libsbml.SpeciesReference
            sid = _check_required(sref, sref.getSpecies(), "species")

            if f_replace and F_SPECIE in f_replace:
                sid = f_replace[F_SPECIE](sid)
            stoichiometry[sid] -= number(
                _check_required(sref, sref.getStoichiometry(), "stoichiometry")
            )

        for (
            sref
        ) in reaction.getListOfProducts():  # noqa: E501 type: libsbml.SpeciesReference
            sid = _check_required(sref, sref.getSpecies(), "species")

            if f_replace and F_SPECIE in f_replace:
                sid = f_replace[F_SPECIE](sid)
            stoichiometry[sid] += number(
                _check_required(sref, sref.getStoichiometry(), "stoichiometry")
            )

        # convert to metabolite objects
        object_stoichiometry = {}
        for met_id in stoichiometry:
            metabolite = cobra_model.metabolites.get_by_id(met_id)
            object_stoichiometry[metabolite] = stoichiometry[met_id]
        cobra_reaction.add_metabolites(object_stoichiometry)

        # GPR
        if r_fbc:
            gpr = None
            gpa = (
                r_fbc.getGeneProductAssociation()
            )  # noqa: E501 type: libsbml.GeneProductAssociation
            if gpa is not None:
                association = (
                    gpa.getAssociation()
                )  # noqa: E501 type: libsbml.FbcAssociation
                gpr = Module(process_association(association))
            cobra_reaction.gpr = GPR(gpr_from=gpr)
        else:
            # fallback to notes information
            notes = cobra_reaction.notes
            if "GENE ASSOCIATION" in notes:
                gpr = notes["GENE ASSOCIATION"]
            elif "GENE_ASSOCIATION" in notes:
                gpr = notes["GENE_ASSOCIATION"]
            else:
                gpr = ""

            if len(gpr) > 0:
                LOGGER.warning(
                    f"Use of GENE ASSOCIATION or GENE_ASSOCIATION "
                    f"in the notes element is discouraged, use "
                    f"fbc:gpr instead: {reaction}"
                )
                if f_replace and F_GENE in f_replace:
                    gpr = " ".join(f_replace[F_GENE](t) for t in gpr.split(" "))
            cobra_reaction.gpr = GPR.from_string(gpr)

    cobra_model.add_reactions(reactions)

    # Objective
    obj_direction = "max"
    coefficients = {}
    if model_fbc:
        obj_list = (
            model_fbc.getListOfObjectives()
        )  # noqa: E501 type: libsbml.ListOfObjectives
        if obj_list is None:
            LOGGER.warning("listOfObjectives element not found")
        elif obj_list.size() == 0:
            LOGGER.warning("No objective in listOfObjectives")
        elif not obj_list.getActiveObjective():
            LOGGER.warning("No active objective in listOfObjectives")
        else:
            obj_id = obj_list.getActiveObjective()
            obj: "libsbml.Objective" = model_fbc.getObjective(obj_id)
            obj_direction = LONG_SHORT_DIRECTION[obj.getType()]

            for (
                flux_obj
            ) in (
                obj.getListOfFluxObjectives()
            ):  # noqa: E501 type: libsbml.FluxObjective
                rid = flux_obj.getReaction()
                if f_replace and F_REACTION in f_replace:
                    rid = f_replace[F_REACTION](rid)
                try:
                    objective_reaction = cobra_model.reactions.get_by_id(rid)
                except KeyError:
                    raise CobraSBMLError(f"Objective reaction '{rid}' not found")
                try:
                    coefficients[objective_reaction] = number(flux_obj.getCoefficient())
                except ValueError as e:
                    LOGGER.warning(str(e))
    else:
        # some legacy models encode objective coefficients in kinetic laws
        reaction: "libsbml.Reaction"
        for reaction in model.getListOfReactions():
            if reaction.isSetKineticLaw():
                klaw: "libsbml.KineticLaw" = reaction.getKineticLaw()
                p_oc: "libsbml.LocalParameter" = klaw.getParameter(
                    "OBJECTIVE_COEFFICIENT"
                )
                if p_oc:
                    rid = _check_required(reaction, reaction.getIdAttribute(), "id")
                    if f_replace and F_REACTION in f_replace:
                        rid = f_replace[F_REACTION](rid)
                    try:
                        objective_reaction = cobra_model.reactions.get_by_id(rid)
                    except KeyError:
                        raise CobraSBMLError(f"Objective reaction '{rid}' not found")
                    try:
                        coefficients[objective_reaction] = number(p_oc.getValue())
                    except ValueError as e:
                        LOGGER.warning(str(e))

                    LOGGER.warning(
                        "Encoding OBJECTIVE_COEFFICIENT in KineticLaw is discouraged, "
                        f"use fbc:fluxObjective instead: {reaction}"
                    )

    if len(coefficients) == 0:
        LOGGER.error(
            "No objective coefficients in model. Unclear what should be optimized"
        )
    set_objective(cobra_model, coefficients)
    cobra_model.solver.objective.direction = obj_direction

    # parse groups
    model_groups: "libsbml.GroupsModelPlugin" = model.getPlugin("groups")
    groups = []
    if model_groups:
        # calculate hashmaps to lookup objects in O(1)
        sid_map = {}
        metaid_map = {}
        for obj_list in [
            model.getListOfCompartments(),
            model.getListOfSpecies(),
            model.getListOfReactions(),
            model_groups.getListOfGroups(),
        ]:
            sbase: "libsbml.SBase"
            for sbase in obj_list:
                if sbase.isSetId():
                    sid_map[sbase.getIdAttribute()] = sbase
                if sbase.isSetMetaId():
                    metaid_map[sbase.getMetaId()] = sbase

        # create groups
        group: "libsbml.Group"
        for group in model_groups.getListOfGroups():
            gid = _check_required(group, group.getIdAttribute(), "id")
            if f_replace and F_GROUP in f_replace:
                gid = f_replace[F_GROUP](gid)
            cobra_group = Group(gid)
            cobra_group.name = group.getName()
            if group.isSetKind():
                cobra_group.kind = group.getKindAsString()
            cobra_group.annotation = _parse_annotations(group)
            cobra_group.notes = _parse_notes_dict(group)

            cobra_members = []
            member: "libsbml.Member"
            for member in group.getListOfMembers():
                if member.isSetIdRef():
                    obj = sid_map[member.getIdRef()]
                elif member.isSetMetaIdRef():
                    obj = metaid_map[member.getMetaIdRef()]

                # noinspection PyUnboundLocalVariable
                typecode = obj.getTypeCode()
                obj_id = _check_required(obj, obj.getIdAttribute(), "id")

                # id replacements
                cobra_member = None
                if typecode == libsbml.SBML_SPECIES:
                    if f_replace and F_SPECIE in f_replace:
                        obj_id = f_replace[F_SPECIE](obj_id)
                    cobra_member = cobra_model.metabolites.get_by_id(obj_id)
                elif typecode == libsbml.SBML_REACTION:
                    if f_replace and F_REACTION in f_replace:
                        obj_id = f_replace[F_REACTION](obj_id)
                    cobra_member = cobra_model.reactions.get_by_id(obj_id)
                    cobra_member.subsystem = group.name
                elif typecode == libsbml.SBML_FBC_GENEPRODUCT:
                    if f_replace and F_GENE in f_replace:
                        obj_id = f_replace[F_GENE](obj_id)
                    cobra_member = cobra_model.genes.get_by_id(obj_id)
                else:
                    LOGGER.warning(
                        f"Member {member} could not be added to group {group}."
                        f"unsupported type code: {typecode}"
                    )

                if cobra_member:
                    cobra_members.append(cobra_member)

            cobra_group.add_members(cobra_members)
            groups.append(cobra_group)
    else:
        # parse deprecated subsystems on reactions
        groups_dict = {}
        for cobra_reaction in cobra_model.reactions:
            if "SUBSYSTEM" in cobra_reaction.notes:
                g_name = cobra_reaction.notes["SUBSYSTEM"]
                cobra_reaction.subsystem = g_name
                if g_name in groups_dict:
                    groups_dict[g_name].append(cobra_reaction)
                else:
                    groups_dict[g_name] = [cobra_reaction]

        for gid, cobra_members in groups_dict.items():
            if f_replace and F_GROUP in f_replace:
                gid = f_replace[F_GROUP](gid)
            cobra_group = Group(gid, name=gid, kind="partonomy")
            cobra_group.annotation["sbo"] = "SBO:0000633"
            cobra_group.add_members(cobra_members)
            groups.append(cobra_group)

    cobra_model.add_groups(groups)

    # general hint for missing flux bounds
    if missing_bounds:
        LOGGER.warning(
            "Missing flux bounds on reactions set to default bounds."
            "As best practise and to avoid confusion flux bounds "
            "should be set explicitly on all reactions."
        )

    return cobra_model


# -----------------------------------------------------------------------------
# Write SBML
# -----------------------------------------------------------------------------
# noinspection PyDefaultArgument
def write_sbml_model(
    cobra_model: Model,
    filename: Union[str, IO, Path],
    f_replace: dict = F_REPLACE,
    **kwargs,
) -> None:
    """Write cobra model to filename.

    The created model is SBML level 3 version 1 (L1V3) with
    fbc package v2 (fbc-v2).

    If the given filename ends with the suffix ".gz" (for example,
    "myfile.xml.gz"), libSBML assumes the caller wants the file to be
    written compressed in gzip format. Similarly, if the given filename
    ends with ".zip" or ".bz2", libSBML assumes the caller wants the
    file to be compressed in zip or bzip2 format (respectively). Files
    whose names lack these suffixes will be written uncompressed. Special
    considerations for the zip format: If the given filename ends with
    ".zip", the file placed in the zip archive will have the suffix
    ".xml" or ".sbml".  For example, the file in the zip archive will
    be named "test.xml" if the given filename is "test.xml.zip" or
    "test.zip". Similarly, the filename in the archive will be
    "test.sbml" if the given filename is "test.sbml.zip".

    Parameters
    ----------
    cobra_model : cobra.core.Model
        Model instance which is written to SBML
    filename : string or filehandle
        path to which the model is written
    f_replace: dict
        dictionary of replacement functions for id replacement
    **kwargs:
        Further keyword arguments are passed on.
    """
    doc = _model_to_sbml(cobra_model, f_replace=f_replace, **kwargs)

    if isinstance(filename, str):
        # write to path
        libsbml.writeSBMLToFile(doc, filename)
    elif isinstance(filename, Path):
        libsbml.writeSBMLToFile(doc, str(filename))
    elif hasattr(filename, "write"):
        # write to file handle
        sbml_str = libsbml.writeSBMLToString(doc)
        filename.write(sbml_str)


def _model_to_sbml(
    cobra_model: Model, f_replace: Optional[dict] = None, units: bool = True
) -> libsbml.SBMLDocument:
    """Convert Cobra model to SBMLDocument.

    Parameters
    ----------
    cobra_model : cobra.core.Model
        Cobra model instance
    f_replace : dict of replacement functions
        Replacement to apply on identifiers.
    units : boolean
        Should the FLUX_UNITS be written in the SBMLDocument.

    Returns
    -------
    libsbml.SBMLDocument
    """
    if f_replace is None:
        f_replace = {}

    sbml_ns = libsbml.SBMLNamespaces(3, 1)  # SBML L3V1
    sbml_ns.addPackageNamespace("fbc", 2)  # fbc-v2

    doc: "libsbml.SBMLDocument" = libsbml.SBMLDocument(sbml_ns)
    doc.setPackageRequired("fbc", False)
    doc.setSBOTerm(SBO_FBA_FRAMEWORK)

    model: "libsbml.Model" = doc.createModel()
    model_fbc: "libsbml.FbcModelPlugin" = model.getPlugin("fbc")
    model_fbc.setStrict(True)

    if cobra_model.id is not None:
        model.setId(cobra_model.id)
        model.setMetaId("meta_" + cobra_model.id)
    else:
        model.setMetaId("meta_model")
    if cobra_model.name is not None:
        model.setName(cobra_model.name)

    # for parsing annotation corresponding to the model
    _sbase_annotations(model, cobra_model.annotation)
    # for parsing notes corresponding to the model
    _sbase_notes_dict(model, cobra_model.notes)

    # Meta information (ModelHistory) related to SBMLDocument
    if hasattr(cobra_model, "_sbml"):
        meta = cobra_model._sbml
        if "annotation" in meta:
            _sbase_annotations(doc, meta["annotation"])
        if "notes" in meta:
            _sbase_notes_dict(doc, meta["notes"])

        history: "libsbml.ModelHistory" = libsbml.ModelHistory()
        if "created" in meta and meta["created"]:
            history.setCreatedDate(libsbml.Date(meta["created"]))
        else:
            time = datetime.datetime.now()
            timestr = time.strftime("%Y-%m-%dT%H:%M:%S")
            date = libsbml.Date(timestr)
            _check(history.setCreatedDate(date), "set creation date")
            _check(history.setModifiedDate(date), "set modified date")

        if "creators" in meta:
            for cobra_creator in meta[
                "creators"
            ]:  # noqa: E501 type: libsbml.ModelCreator
                creator = libsbml.ModelCreator()
                if cobra_creator.get("familyName", None):
                    creator.setFamilyName(cobra_creator["familyName"])
                if cobra_creator.get("givenName", None):
                    creator.setGivenName(cobra_creator["givenName"])
                if cobra_creator.get("organisation", None):
                    creator.setOrganisation(cobra_creator["organisation"])
                if cobra_creator.get("email", None):
                    creator.setEmail(cobra_creator["email"])

                _check(history.addCreator(creator), "adding creator to ModelHistory.")

        # TODO: Will be implemented as part of
        #  https://github.com/opencobra/cobrapy/issues/810
        # _check(model.setModelHistory(history), 'set model history')

    # Units
    flux_udef = None
    if units:
        flux_udef: "libsbml.UnitDefinition" = model.createUnitDefinition()
        flux_udef.setId(UNITS_FLUX[0])
        for u in UNITS_FLUX[1]:
            unit: "libsbml.Unit" = flux_udef.createUnit()
            unit.setKind(u.kind)
            unit.setExponent(u.exponent)
            unit.setScale(u.scale)
            unit.setMultiplier(u.multiplier)

    min_value = config.lower_bound
    max_value = config.upper_bound

    _create_parameter(
        model, pid=LOWER_BOUND_ID, value=min_value, sbo=SBO_DEFAULT_FLUX_BOUND
    )
    _create_parameter(
        model, pid=UPPER_BOUND_ID, value=max_value, sbo=SBO_DEFAULT_FLUX_BOUND
    )
    _create_parameter(model, pid=ZERO_BOUND_ID, value=0, sbo=SBO_DEFAULT_FLUX_BOUND)
    _create_parameter(
        model, pid=BOUND_MINUS_INF, value=-float("Inf"), sbo=SBO_DEFAULT_FLUX_BOUND
    )
    _create_parameter(
        model, pid=BOUND_PLUS_INF, value=float("Inf"), sbo=SBO_DEFAULT_FLUX_BOUND
    )

    # Compartments
    # FIXME: use first class compartment model (and write notes & annotations)
    #     (https://github.com/opencobra/cobrapy/issues/811)
    for cid, name in cobra_model.compartments.items():
        compartment: "libsbml.Compartment" = model.createCompartment()
        compartment.setId(cid)
        compartment.setName(name)
        compartment.setConstant(True)

        # FIXME: write annotations and notes
        # _sbase_notes(c, com.notes)
        # _sbase_annotations(c, com.annotation)

    # Species
    for metabolite in cobra_model.metabolites:
        specie: "libsbml.Species" = model.createSpecies()
        mid = metabolite.id
        if f_replace and F_SPECIE_REV in f_replace:
            mid = f_replace[F_SPECIE_REV](mid)
        specie.setId(mid)
        specie.setConstant(False)
        specie.setBoundaryCondition(False)
        specie.setHasOnlySubstanceUnits(False)
        specie.setName(metabolite.name)
        specie.setCompartment(metabolite.compartment)
        s_fbc: "libsbml.FbcSpeciesPlugin" = specie.getPlugin("fbc")
        if metabolite.charge is not None:
            s_fbc.setCharge(metabolite.charge)
        if metabolite.formula is not None:
            s_fbc.setChemicalFormula(metabolite.formula)

        _sbase_annotations(specie, metabolite.annotation)
        _sbase_notes_dict(specie, metabolite.notes)

    # Genes
    for cobra_gene in cobra_model.genes:
        gp: "libsbml.GeneProduct" = model_fbc.createGeneProduct()
        gid = cobra_gene.id
        if f_replace and F_GENE_REV in f_replace:
            gid = f_replace[F_GENE_REV](gid)
        gp.setId(gid)
        gname = cobra_gene.name
        if gname is None or len(gname) == 0:
            gname = gid
        gp.setName(gname)
        gp.setLabel(gid)

        _sbase_annotations(gp, cobra_gene.annotation)
        _sbase_notes_dict(gp, cobra_gene.notes)

    # Objective
    objective: "libsbml.Objective" = model_fbc.createObjective()
    objective.setId("obj")
    objective.setType(SHORT_LONG_DIRECTION[cobra_model.objective.direction])
    model_fbc.setActiveObjectiveId("obj")

    # Reactions
    reaction_coefficients = linear_reaction_coefficients(cobra_model)
    for cobra_reaction in cobra_model.reactions:
        rid = cobra_reaction.id
        if f_replace and F_REACTION_REV in f_replace:
            rid = f_replace[F_REACTION_REV](rid)
        reaction: "libsbml.Reaction" = model.createReaction()
        reaction.setId(rid)
        reaction.setName(cobra_reaction.name)
        reaction.setFast(False)
        reaction.setReversible((cobra_reaction.lower_bound < 0))
        _sbase_annotations(reaction, cobra_reaction.annotation)
        _sbase_notes_dict(reaction, cobra_reaction.notes)

        # stoichiometry
        for metabolite, stoichiometry in cobra_reaction.metabolites.items():
            sid = metabolite.id
            if f_replace and F_SPECIE_REV in f_replace:
                sid = f_replace[F_SPECIE_REV](sid)
            if stoichiometry < 0:
                sref = (
                    reaction.createReactant()
                )  # noqa: E501 type: libsbml.SpeciesReference
                sref.setSpecies(sid)
                sref.setStoichiometry(-stoichiometry)
                sref.setConstant(True)
            else:
                sref = (
                    reaction.createProduct()
                )  # noqa: E501 type: libsbml.SpeciesReference
                sref.setSpecies(sid)
                sref.setStoichiometry(stoichiometry)
                sref.setConstant(True)

        # bounds
        r_fbc: "libsbml.FbcReactionPlugin" = reaction.getPlugin("fbc")
        r_fbc.setLowerFluxBound(
            _create_bound(
                model,
                cobra_reaction,
                "lower_bound",
                f_replace=f_replace,
                units=units,
                flux_udef=flux_udef,
            )
        )
        r_fbc.setUpperFluxBound(
            _create_bound(
                model,
                cobra_reaction,
                "upper_bound",
                f_replace=f_replace,
                units=units,
                flux_udef=flux_udef,
            )
        )

        # GPR
        gpr = cobra_reaction.gpr
        if gpr is not None and gpr.body:
            # replace ids in string
            if f_replace and F_GENE_REV in f_replace:
                idmap = {gid: f_replace[F_GENE_REV](gid) for gid in gpr.genes}
                gpr_new = gpr.to_string(names=idmap)
            else:
                gpr_new = gpr.to_string()

            gpa = (
                r_fbc.createGeneProductAssociation()
            )  # noqa: E501 type: libsbml.GeneProductAssociation
            # uses ids to identify GeneProducts (True),
            # does not create GeneProducts (False)
            _check(gpa.setAssociation(gpr_new, True, False), "set gpr: " + gpr_new)

        # objective coefficients
        if reaction_coefficients.get(cobra_reaction, 0) != 0:
            flux_obj = (
                objective.createFluxObjective()
            )  # noqa: E501 type: libsbml.FluxObjective
            flux_obj.setReaction(rid)
            flux_obj.setCoefficient(cobra_reaction.objective_coefficient)

    # write groups
    if len(cobra_model.groups) > 0:
        doc.enablePackage(
            "http://www.sbml.org/sbml/level3/version1/groups/version1", "groups", True
        )
        doc.setPackageRequired("groups", False)
        model_group = model.getPlugin(
            "groups"
        )  # noqa: E501 type: libsbml.GroupsModelPlugin
        for cobra_group in cobra_model.groups:
            group: "libsbml.Group" = model_group.createGroup()
            if f_replace and F_GROUP_REV in f_replace:
                gid = f_replace[F_GROUP_REV](cobra_group.id)
            else:
                gid = cobra_group.id
            group.setId(gid)
            group.setName(cobra_group.name)
            group.setKind(cobra_group.kind)

            _sbase_notes_dict(group, cobra_group.notes)
            _sbase_annotations(group, cobra_group.annotation)

            for cobra_member in cobra_group.members:
                member: "libsbml.Member" = group.createMember()
                mid = cobra_member.id
                m_type = str(type(cobra_member))

                # id replacements
                if "Reaction" in m_type:
                    if f_replace and F_REACTION_REV in f_replace:
                        mid = f_replace[F_REACTION_REV](mid)
                if "Metabolite" in m_type:
                    if f_replace and F_SPECIE_REV in f_replace:
                        mid = f_replace[F_SPECIE_REV](mid)
                if "Gene" in m_type:
                    if f_replace and F_GENE_REV in f_replace:
                        mid = f_replace[F_GENE_REV](mid)

                member.setIdRef(mid)
                if cobra_member.name and len(cobra_member.name) > 0:
                    member.setName(cobra_member.name)

    return doc


def _create_bound(
    model: libsbml.Model,
    reaction: Reaction,
    bound_type: str,
    f_replace: dict,
    units: Optional[bool] = None,
    flux_udef: Optional[libsbml.UnitDefinition] = None,
) -> str:
    """Create bound in model for given reaction.

    Adds the parameters for the bounds to the SBML model.

    Parameters
    ----------
    model : libsbml.Model
        SBML model instance
    reaction : cobra.core.Reaction
        Cobra reaction instance from which the bounds are read.
    bound_type : {LOWER_BOUND, UPPER_BOUND}
        Type of bound
    f_replace : dict
        of id replacement functions
    units : bool, optional, defualt None
        Whether or not to use flux units in the SBML document.
    flux_udef: libsbml.UnitDefinition, optional
        Unit definition if units are used.

    Returns
    -------
    pid: str
        Id of bound parameter.
    """
    value = getattr(reaction, bound_type)
    if value == config.lower_bound:
        return LOWER_BOUND_ID
    elif value == 0:
        return ZERO_BOUND_ID
    elif value == config.upper_bound:
        return UPPER_BOUND_ID
    elif value == -float("Inf"):
        return BOUND_MINUS_INF
    elif value == float("Inf"):
        return BOUND_PLUS_INF
    else:
        # new parameter
        rid = reaction.id
        if f_replace and F_REACTION_REV in f_replace:
            rid = f_replace[F_REACTION_REV](rid)
        pid = rid + "_" + bound_type
        _create_parameter(
            model,
            pid=pid,
            value=value,
            sbo=SBO_FLUX_BOUND,
            units=units,
            flux_udef=flux_udef,
        )
        return pid


def _create_parameter(
    model: libsbml.Model,
    pid: str,
    value: float,
    sbo: Optional[str] = None,
    constant: Optional[bool] = True,
    units: Optional[bool] = None,
    flux_udef: Optional[libsbml.UnitDefinition] = None,
) -> None:
    """Create parameter in SBML model.

    Parameters
    ----------
    model : libsbml.Model
        SBML model instance
    pid : str
        Parameter id to create in the SBML model.
    value: float
        Value to set parameter
    sbo: str, optional
        SBO term for parameter. In COBRA, it seems to be SBO_FLUX_BOUND ( "SBO:0000625")
        or SBO_DEFAULT_FLUX_BOUND ("SBO:0000626")
    constant: bool, optional
        Flag if parameter is constant.
    units : bool, optional, defualt None
        Whether or not to use flux units in the SBML document.
    flux_udef: libsbml.UnitDefinition, optional
        Unit definition if units are used.

    """
    parameter: "libsbml.Parameter" = model.createParameter()
    parameter.setId(pid)
    parameter.setValue(value)
    parameter.setConstant(constant)
    if sbo:
        parameter.setSBOTerm(sbo)
    if units:
        parameter.setUnits(flux_udef.getId())


def _check_required(sbase: "libsbml.Base", value: str, attribute: str) -> str:
    """Get required attribute from SBase.

    Parameters
    ----------
    sbase : libsbml.SBase
    value : existing value
    attribute: name of attribute

    Returns
    -------
    attribute value (or value if already set)

    Raises
    ------
    CobraSBMLError if attribute not found or not parsed.
    """
    if (value is None) or (value == ""):
        msg = (
            f"Required attribute '{attribute}' cannot be "
            f"found or parsed in '{sbase}'."
        )
        if hasattr(sbase, "getId") and sbase.getId():
            msg += f" with id '{sbase.getId()}'"
        elif hasattr(sbase, "getName") and sbase.getName():
            msg += f" with name '{sbase.getName()}'"
        elif hasattr(sbase, "getMetaId") and sbase.getMetaId():
            msg += f" with metaId '{sbase.getName()}'"
        raise CobraSBMLError(msg)
    if attribute == "id" and not libsbml.SyntaxChecker.isValidSBMLSId(value):
        LOGGER.error(f"'{value}' is not a valid SBML 'SId'.")

    return value


def _check(value: Union[None, int], message: str) -> None:
    """
    Check the libsbml return value and logs error messages.

    Parameters
    ----------
    value: None or int
    message: str

    If 'value' is None, logs an error message constructed using
      'message' and then exits with status code 1. If 'value' is an integer,
      it assumes it is a libSBML return status code. If the code value is
      LIBSBML_OPERATION_SUCCESS, returns without further action; if it is not,
      prints an error message constructed using 'message' along with text from
      libSBML explaining the meaning of the code, and exits with status code 1.

    """
    if value is None:
        LOGGER.error(
            "Error: LibSBML returned a null value trying to <" + message + ">."
        )
    elif type(value) is int:
        if value == libsbml.LIBSBML_OPERATION_SUCCESS:
            return
        else:
            LOGGER.error("Error encountered trying to <" + message + ">.")
            LOGGER.error(
                f"LibSBML error code {str(value)}: "
                f"{libsbml.OperationReturnValue_toString(value).strip()}"
            )
    else:
        return


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------
def _parse_notes_dict(sbase) -> dict:
    """Create dictionary of COBRA notes.

    Parameters
    ----------
    sbase : libsbml.SBase

    Returns
    -------
    dict of notes
    """
    notes = sbase.getNotesString()
    if notes and len(notes) > 0:
        notes_store = {}
        for match in pattern_notes.finditer(notes):
            _content = match.group("content")
            try:
                # Python 2.7 does not allow keywords for split.
                # Python 3 can have (":", maxsplit=1)
                key, value = _content.split(":", maxsplit=1)
            except ValueError:
                LOGGER.debug(f"Unexpected content format '{_content}'.")
                continue
            notes_store[key.strip()] = value.strip()
        return {k: v for k, v in notes_store.items() if len(v) > 0}
    else:
        return {}


def _sbase_notes_dict(sbase: libsbml.SBase, notes: dict) -> None:
    """Set SBase notes based on dictionary.

    Parameters
    ----------
    sbase : libsbml.SBase
        SBML object to set notes on
    notes : dict
        Notes information from cobra object.
    """
    if notes and len(notes) > 0:
        tokens = (
            ['<html xmlns = "http://www.w3.org/1999/xhtml" >']
            + [f"<p>{k}: {v}</p>" for (k, v) in notes.items()]
            + ["</html>"]
        )
        _check(
            sbase.setNotes("\n".join(tokens)),
            f"Setting notes on sbase: {sbase}",
        )


# -----------------------------------------------------------------------------
# Annotations
# -----------------------------------------------------------------------------
"""
cobra annotations will be dictionaries of the form:
    object.annotation = {
        'provider' : [(qualifier, entity), ...]
    }
A concrete example for a metabolite would look like the following
    metabolite.annotation = {
        'chebi': [(isVersionOf, "CHEBI:17234), (is, "CHEBI:4167),],
        'kegg.compound': [(is, "C00031")]
    }
The providers are hereby MIRIAM registry keys for collections
https://www.ebi.ac.uk/miriam/main/collections
The qualifiers are biomodel qualifiers
https://co.mbine.org/standards/qualifiers

In the current stage the new annotation format is not completely supported yet.
"""

URL_IDENTIFIERS_PATTERN = re.compile(r"^https?://identifiers.org/(.+?)[:/](.+)")

URL_IDENTIFIERS_PREFIX = "https://identifiers.org"
QUALIFIER_TYPES = {
    "is": libsbml.BQB_IS,
    "hasPart": libsbml.BQB_HAS_PART,
    "isPartOf": libsbml.BQB_IS_PART_OF,
    "isVersionOf": libsbml.BQB_IS_VERSION_OF,
    "hasVersion": libsbml.BQB_HAS_VERSION,
    "isHomologTo": libsbml.BQB_IS_HOMOLOG_TO,
    "isDescribedBy": libsbml.BQB_IS_DESCRIBED_BY,
    "isEncodedBy": libsbml.BQB_IS_ENCODED_BY,
    "encodes": libsbml.BQB_ENCODES,
    "occursIn": libsbml.BQB_OCCURS_IN,
    "hasProperty": libsbml.BQB_HAS_PROPERTY,
    "isPropertyOf": libsbml.BQB_IS_PROPERTY_OF,
    "hasTaxon": libsbml.BQB_HAS_TAXON,
    "unknown": libsbml.BQB_UNKNOWN,
    "bqm_is": libsbml.BQM_IS,
    "bqm_isDescribedBy": libsbml.BQM_IS_DESCRIBED_BY,
    "bqm_isDerivedFrom": libsbml.BQM_IS_DERIVED_FROM,
    "bqm_isInstanceOf": libsbml.BQM_IS_INSTANCE_OF,
    "bqm_hasInstance": libsbml.BQM_HAS_INSTANCE,
    "bqm_unknown": libsbml.BQM_UNKNOWN,
}


def _parse_annotations(sbase: libsbml.SBase) -> dict:
    """Parse cobra annotations from a given SBase object.

    Annotations are dictionaries with the providers as keys.

    Parameters
    ----------
    sbase : libsbml.SBase
        SBase from which the SBML annotations are read

    Returns
    -------
    dict (annotation dictionary)
    """
    #     FIXME: annotation format must be updated (this is a big collection of
    #           fixes) - see: https://github.com/opencobra/cobrapy/issues/684)
    annotation = {}

    # SBO term
    if sbase.isSetSBOTerm():
        # FIXME: correct handling of annotations
        annotation["sbo"] = sbase.getSBOTermID()

    # RDF annotation
    cvterms = sbase.getCVTerms()
    if cvterms is None:
        return annotation

    cvterm: "libsbml.CVTerm"
    for cvterm in cvterms:
        for k in range(cvterm.getNumResources()):
            # FIXME: read and store the qualifier

            uri = cvterm.getResourceURI(k)
            data = _parse_annotation_info(uri)
            if data is None:
                continue
            else:
                provider, identifier = data

            if provider in annotation:
                if isinstance(annotation[provider], str):
                    annotation[provider] = [annotation[provider]]
                # FIXME: use a list
                if identifier not in annotation[provider]:
                    annotation[provider].append(identifier)
            else:
                # FIXME: always in list
                annotation[provider] = identifier

    return annotation


def _parse_annotation_info(uri: str) -> Union[None, Tuple[str, str]]:
    """Parse provider and term from given identifiers annotation uri.

    Parameters
    ----------
    uri : str
        uri (identifiers.org url)

    Returns
    -------
    (provider, identifier) if resolvable, None otherwise
    """
    match = URL_IDENTIFIERS_PATTERN.match(uri)
    if match:
        provider, identifier = match.group(1), match.group(2)
        if provider.isupper():
            identifier = f"{provider}:{identifier}"
            provider = provider.lower()
    else:
        LOGGER.warning(
            f"{uri} does not conform to "
            f"'http(s)://identifiers.org/collection/id' or"
            f"'http(s)://identifiers.org/COLLECTION:id"
        )
        return None

    return provider, identifier


def _sbase_annotations(sbase: libsbml.SBase, annotation: dict) -> None:
    """Set SBase annotations based on cobra annotations.

    Parameters
    ----------
    sbase : libsbml.SBase
        SBML object to annotate
    annotation : dict, cobra annotation structure
        cobra object with annotation information

    Raises
    ------
    CobraSBMLError for unsupported qualifier
    """
    #    FIXME: annotation format must be updated
    #     (https://github.com/opencobra/cobrapy/issues/684)
    if not annotation or len(annotation) == 0:
        return

    # standardize annotations
    annotation_data = deepcopy(annotation)

    for key, value in annotation_data.items():
        # handling of non-string annotations (e.g. integers)
        if isinstance(value, (float, int)):
            value = str(value)
        if isinstance(value, str):
            annotation_data[key] = [("is", value)]

    for _key, value in annotation_data.items():
        for idx, item in enumerate(value):
            if isinstance(item, str):
                value[idx] = ("is", item)

    # set metaId
    meta_id = f"meta_{sbase.getId()}"
    sbase.setMetaId(meta_id)

    # rdf_items = []
    for provider, data in annotation_data.items():
        # set SBOTerm
        if provider in ["SBO", "sbo"]:
            if provider == "SBO":
                LOGGER.warning(
                    "'SBO' provider is deprecated, use 'sbo' provider instead"
                )
            sbo_term = data[0][1]
            _check(sbase.setSBOTerm(sbo_term), f"Setting SBOTerm: {sbo_term}")

            # FIXME: sbo should also be written as CVTerm
            continue

        for item in data:
            qualifier_str, entity = item[0], item[1]
            qualifier = QUALIFIER_TYPES.get(qualifier_str, None)
            if qualifier is None:
                qualifier = libsbml.BQB_IS
                LOGGER.error(
                    f"Qualifier type is not supported on annotation: '{qualifier_str}'"
                )

            qualifier_type = libsbml.BIOLOGICAL_QUALIFIER
            if qualifier_str.startswith("bqm_"):
                qualifier_type = libsbml.MODEL_QUALIFIER

            cv: "libsbml.CVTerm" = libsbml.CVTerm()
            cv.setQualifierType(qualifier_type)
            if qualifier_type == libsbml.BIOLOGICAL_QUALIFIER:
                cv.setBiologicalQualifierType(qualifier)
            elif qualifier_type == libsbml.MODEL_QUALIFIER:
                cv.setModelQualifierType(qualifier)
            else:
                raise CobraSBMLError(f"Unsupported qualifier: {qualifier}")
            resource = f"{URL_IDENTIFIERS_PREFIX}/{provider}/{entity}"
            cv.addResource(resource)
            _check(
                sbase.addCVTerm(cv),
                f"Setting cvterm: {cv}, resource: {resource}",
            )


# -----------------------------------------------------------------------------
# Validation
# -----------------------------------------------------------------------------
def validate_sbml_model(
    filename: Union[str, IO, Path],
    check_model: bool = True,
    internal_consistency: bool = True,
    check_units_consistency: bool = False,
    check_modeling_practice: bool = False,
    **kwargs,
) -> Tuple[Optional[Model], dict]:
    """Validate SBML model and returns the model along with a list of errors.

    Parameters
    ----------
    filename : str or filehandle
        The filename (or SBML string) of the SBML model to be validated.
    check_model: boolean {True, False}, default True
        Whether to also check some basic model properties such as reaction
        boundaries and compartment formulas.
    internal_consistency: bool, optional
        Check internal consistency (default True).
    check_units_consistency: bool, optional
        Check consistency of units (default True).
    check_modeling_practice: bool, optional
        Check modeling practise (defualt True).
    **kwargs:
        Further keyword arguments are passed on to the called function (_sbml_to_model).

    Returns
    -------
    (model, errors)
    model : :class:`~cobra.core.Model.Model` object
        The cobra model if the file could be read successfully or None
        otherwise.
    errors : dict
        Warnings and errors grouped by their respective types.

    Notes
    -----
    Errors and warnings are grouped based on their type. SBML_* types are
    from the libsbml validator. COBRA_* types are from the cobrapy SBML
    parser.
    """
    keys = (
        "SBML_FATAL",
        "SBML_ERROR",
        "SBML_SCHEMA_ERROR",
        "SBML_WARNING",
        "COBRA_FATAL",
        "COBRA_ERROR",
        "COBRA_WARNING",
        "COBRA_CHECK",
    )
    errors = {key: [] for key in keys}

    # [1] libsbml validation
    doc: "libsbml.SBMLDocument" = _get_doc_from_filename(filename)

    # set checking of units & modeling practise
    doc.setConsistencyChecks(
        libsbml.LIBSBML_CAT_UNITS_CONSISTENCY, check_units_consistency
    )
    doc.setConsistencyChecks(
        libsbml.LIBSBML_CAT_MODELING_PRACTICE, check_modeling_practice
    )

    # check internal consistency
    if internal_consistency:
        doc.checkInternalConsistency()
    doc.checkConsistency()

    for k in range(doc.getNumErrors()):
        e: "libsbml.SBMLError" = doc.getError(k)
        msg = _error_string(e, k=k)
        sev = e.getSeverity()
        if sev == libsbml.LIBSBML_SEV_FATAL:
            errors["SBML_FATAL"].append(msg)
        elif sev == libsbml.LIBSBML_SEV_ERROR:
            errors["SBML_ERROR"].append(msg)
        elif sev == libsbml.LIBSBML_SEV_SCHEMA_ERROR:
            errors["SBML_SCHEMA_ERROR"].append(msg)
        elif sev == libsbml.LIBSBML_SEV_WARNING:
            errors["SBML_WARNING"].append(msg)

    # [2] cobrapy validation (check that SBML can be read into model)
    # all warnings generated while loading will be logged as errors
    log_stream = StringIO()
    stream_handler = logging.StreamHandler(log_stream)
    formatter = logging.Formatter("%(levelname)s:%(message)s")
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)
    LOGGER.addHandler(stream_handler)
    LOGGER.propagate = False

    try:
        # read model and allow additional parser arguments
        model = _sbml_to_model(doc, **kwargs)
    except CobraSBMLError as e:
        errors["COBRA_ERROR"].append(str(e))
        return None, errors
    except Exception as e:
        errors["COBRA_FATAL"].append(str(e))
        return None, errors

    cobra_errors = log_stream.getvalue().split("\n")
    for cobra_error in cobra_errors:
        tokens = cobra_error.split(":")
        error_type = tokens[0]
        error_msg = ":".join(tokens[1:])

        if error_type == "WARNING":
            errors["COBRA_WARNING"].append(error_msg)
        elif error_type == "ERROR":
            errors["COBRA_ERROR"].append(error_msg)

    # remove stream handler
    LOGGER.removeHandler(stream_handler)
    LOGGER.propagate = True

    # [3] additional model tests
    if check_model:
        errors["COBRA_CHECK"].extend(check_metabolite_compartment_formula(model))

    for key in ["SBML_FATAL", "SBML_ERROR", "SBML_SCHEMA_ERROR"]:
        if len(errors[key]) > 0:
            LOGGER.error("SBML errors in validation, check error log for details.")
            break
    for key in ["SBML_WARNING"]:
        if len(errors[key]) > 0:
            LOGGER.error("SBML warnings in validation, check error log for details.")
            break
    for key in ["COBRA_FATAL", "COBRA_ERROR"]:
        if len(errors[key]) > 0:
            LOGGER.error("COBRA errors in validation, check error log for details.")
            break
    for key in ["COBRA_WARNING", "COBRA_CHECK"]:
        if len(errors[key]) > 0:
            LOGGER.error("COBRA warnings in validation, check error log for details.")
            break

    return model, errors


def _error_string(error: "libsbml.SBMLError", k: Optional[int] = None):
    """Return string representation of SBMLError.

    Parameters
    ----------
    error : libsbml.SBMLError
    k : int, optional
        index of error (default None).

    Returns
    -------
    string representation of error
    """
    package = error.getPackage()
    if package == "":
        package = "core"

    error_str = (
        f"E{k} ({error.getSeverityAsString()}): "
        f"{error.getCategoryAsString()} "
        f"({package}, L{error.getLine()}); "
        f"{error.getShortMessage()}; {error.getMessage()}"
    )
    return error_str



"""Provide the OptGP sampler class and helper functions."""


from typing import TYPE_CHECKING, Dict, Optional, Tuple

import numpy as np
import pandas as pd

from ..core.configuration import Configuration
from ..util import ProcessPool
from .core import step
from .hr_sampler import HRSampler, shared_np_array


if TYPE_CHECKING:
    from cobra import Model


__all__ = ("OptGPSampler",)


configuration = Configuration()


class OptGPSampler(HRSampler):
    """
    Improved Artificial Centering Hit-and-Run sampler.

    A parallel sampler with fast convergence and parallel execution.
    See [1]_ for details.

    Parameters
    ----------
    model : cobra.Model
        The cobra model from which to generate samples.
    processes: int, optional
        The number of processes used during sampling
        (default cobra.Configuration.processes).
    thinning : int, optional
        The thinning factor of the generated sampling chain. A thinning of
        10 means samples are returned every 10 steps (default 100).
    nproj : int > 0, optional
        How often to reproject the sampling point into the feasibility
        space. Avoids numerical issues at the cost of lower sampling. If
        you observe many equality constraint violations with
        `sampler.validate` you should lower this number (default None).
    seed : int > 0, optional
        Sets the random number seed. Initialized to the current time stamp
        if None (default None).

    Attributes
    ----------
    n_samples : int
        The total number of samples that have been generated by this
        sampler instance.
    problem : typing.NamedTuple
        A NamedTuple whose attributes define the entire sampling problem in
        matrix form.
    warmup : numpy.matrix
        A numpy matrix with as many columns as reactions in the model and
        more than 3 rows containing a warmup sample in each row. None if no
        warmup points have been generated yet.
    retries : int
        The overall of sampling retries the sampler has observed. Larger
        values indicate numerical instabilities.
    fwd_idx : numpy.array
        A numpy array having one entry for each reaction in the model,
        containing the index of the respective forward variable.
    rev_idx : numpy.array
        A numpy array having one entry for each reaction in the model,
        containing the index of the respective reverse variable.
    prev : numpy.array
        The current/last flux sample generated.
    center : numpy.array
        The center of the sampling space as estimated by the mean of all
        previously generated samples.

    Notes
    -----
    The sampler is very similar to artificial centering where each process
    samples its own chain. Initial points are chosen randomly from the
    warmup points followed by a linear transformation that pulls the points
    a little bit towards the center of the sampling space.

    If the number of processes used is larger than the one requested,
    number of samples is adjusted to the smallest multiple of the number of
    processes larger than the requested sample number. For instance, if you
    have 3 processes and request 8 samples, you will receive 9.

    Memory usage is roughly in the order of (2 * number of reactions)^2
    due to the required nullspace matrices and warmup points. So, large
    models easily take up a few GBs of RAM. However, most of the large
    matrices are kept in shared memory. So the RAM usage is independent of
    the number of processes.

    References
    ----------
    .. [1] Megchelenbrink W, Huynen M, Marchiori E (2014)
       optGpSampler: An Improved Tool for Uniformly Sampling the Solution-Space
       of Genome-Scale Metabolic Networks.
       PLoS ONE 9(2): e86587.
       https://doi.org/10.1371/journal.pone.0086587

    """

    def __init__(
        self,
        model: "Model",
        thinning: int = 100,
        processes: Optional[int] = None,
        nproj: Optional[int] = None,
        seed: Optional[int] = None,
        **kwargs
    ) -> None:
        """Initialize a new OptGPSampler."""
        super().__init__(model, thinning, nproj=nproj, seed=seed, *kwargs)
        self.generate_fva_warmup()

        if processes is None:
            self.processes = configuration.processes
        else:
            self.processes = processes

        # This maps our saved center into shared memory,
        # meaning they are synchronized across processes
        self.center = shared_np_array(
            (len(self.model.variables),), self.warmup.mean(axis=0)
        )

    def sample(self, n: int, fluxes: bool = True) -> pd.DataFrame:
        """Generate a set of samples.

        This is the basic sampling function for all hit-and-run samplers.

        Parameters
        ----------
        n : int
            The minimum number of samples that are generated at once.
        fluxes : bool, optional
            Whether to return fluxes or the internal solver variables. If
            set to False, will return a variable for each forward and
            backward flux as well as all additional variables you might
            have defined in the model (default True).

        Returns
        -------
        pandas.DataFrame
            Returns a pandas DataFrame with `n` rows, each containing a
            flux sample.

        Notes
        -----
        Performance of this function linearly depends on the number
        of reactions in your model and the thinning factor.

        If the number of processes is larger than one, computation is split
        across the CPU cores of your machine. This may shorten computation
        time. However, there is also overhead in setting up parallel
        computation primitives so, we recommend to calculate large numbers
        of samples at once (`n` > 1000).

        """
        if self.processes > 1:
            n_process = np.ceil(n / self.processes).astype(int)
            n = n_process * self.processes

            # The cast to list is weird but not doing it gives recursion
            # limit errors, something weird going on with multiprocessing
            args = list(zip([n_process] * self.processes, range(self.processes)))

            with ProcessPool(
                self.processes, initializer=mp_init, initargs=(self,)
            ) as pool:
                results = pool.map(_sample_chain, args, chunksize=1)

            chains = np.vstack([r[1] for r in results])
            self.retries += sum(r[0] for r in results)
        else:
            mp_init(self)
            results = _sample_chain((n, 0))
            chains = results[1]

        # Update the global center
        self.center = (self.n_samples * self.center + np.atleast_2d(chains).sum(0)) / (
            self.n_samples + n
        )
        self.n_samples += n

        if fluxes:
            names = [r.id for r in self.model.reactions]

            return pd.DataFrame(
                chains[:, self.fwd_idx] - chains[:, self.rev_idx],
                columns=names,
            )
        else:
            names = [v.name for v in self.model.variables]

            return pd.DataFrame(chains, columns=names)

    # Models can be large so don't pass them around during multiprocessing
    def __getstate__(self) -> Dict:
        """Return the object for serialization."""
        d = dict(self.__dict__)
        del d["model"]
        return d


def mp_init(obj: OptGPSampler) -> None:
    """Initialize the multiprocessing pool."""
    global sampler
    sampler = obj


# Unfortunately this has to be outside the class to be usable with
# multiprocessing :()
def _sample_chain(args: Tuple[int, int]) -> Tuple[int, OptGPSampler]:
    """Sample a single chain for OptGPSampler.

    `center` and `n_samples` are updated locally and forgotten afterwards.

    """
    n, idx = args  # has to be this way to work in Python 2.7
    center = sampler.center
    np.random.seed((sampler._seed + idx) % np.iinfo(np.int32).max)
    pi = np.random.randint(sampler.n_warmup)

    prev = sampler.warmup[pi, :]
    prev = step(sampler, center, prev - center, 0.95)

    n_samples = max(sampler.n_samples, 1)
    samples = np.zeros((n, center.shape[0]))

    for i in range(1, sampler.thinning * n + 1):
        pi = np.random.randint(sampler.n_warmup)
        delta = sampler.warmup[pi, :] - center

        prev = step(sampler, prev, delta)

        if sampler.problem.homogeneous and (
            n_samples * sampler.thinning % sampler.nproj == 0
        ):
            prev = sampler._reproject(prev)
            center = sampler._reproject(center)

        if i % sampler.thinning == 0:
            samples[i // sampler.thinning - 1, :] = prev

        center = (n_samples * center) / (n_samples + 1) + prev / (n_samples + 1)
        n_samples += 1

    return (sampler.retries, samples)



"""Provide a wrapper function for performing flux sampling of cobra models."""

from typing import TYPE_CHECKING, Optional

import pandas as pd

from .achr import ACHRSampler
from .optgp import OptGPSampler


if TYPE_CHECKING:
    from cobra import Model


def sample(
    model: "Model",
    n: int,
    method: str = "optgp",
    thinning: int = 100,
    processes: int = 1,
    seed: Optional[int] = None,
) -> pd.DataFrame:
    """Sample valid flux distributions from a cobra model.

    Currently, two methods are supported:

    1. 'optgp' (default) which uses the OptGPSampler that supports parallel
        sampling. Requires large numbers of samples to be performant
        (`n` > 1000). For smaller samples, 'achr' might be better suited.
        For details, refer [1]_ .

    2. 'achr' which uses artificial centering hit-and-run. This is a single
       process method with good convergence. For details, refer [2]_ .

    Parameters
    ----------
    model : cobra.Model
        The model from which to sample flux distributions.
    n : int
        The number of samples to obtain. When using 'optgp', this must be a
        multiple of `processes`, otherwise a larger number of samples will
        be returned.
    method : {"optgp", "achr"}, optional
        The sampling algorithm to use (default "optgp").
    thinning : int, optional
        The thinning factor of the generated sampling chain. A thinning of
        10 means samples are returned every 10 steps. Defaults to 100 which
        in benchmarks gives approximately uncorrelated samples. If set to 1
        will return all iterates (default 100).
    processes : int, optional
        Only used for 'optgp'. The number of processes used to generate
        samples (default 1).
    seed : int > 0, optional
        Sets the random number seed. Initialized to the current time stamp
        if None (default None).

    Returns
    -------
    pandas.DataFrame
        The generated flux samples. Each row corresponds to a sample of the
        fluxes and the columns are the reactions.

    Notes
    -----
    The samplers have a correction method to ensure equality feasibility for
    long-running chains, however this will only work for homogeneous models,
    meaning models with no non-zero fixed variables or constraints (
    right-hand side of the equalities are zero).

    References
    ----------
    .. [1] Megchelenbrink W, Huynen M, Marchiori E (2014)
       optGpSampler: An Improved Tool for Uniformly Sampling the Solution-Space
       of Genome-Scale Metabolic Networks.
       PLoS ONE 9(2): e86587.
       https://doi.org/10.1371/journal.pone.0086587

    .. [2] Direction Choice for Accelerated Convergence in Hit-and-Run Sampling
       David E. Kaufman, Robert L. Smith
       Operations Research 199846:1 , 84-95
       https://doi.org/10.1287/opre.46.1.84

    """
    if method == "optgp":
        sampler = OptGPSampler(model, processes=processes, thinning=thinning, seed=seed)
    elif method == "achr":
        sampler = ACHRSampler(model, thinning=thinning, seed=seed)
    else:
        raise ValueError(
            f'Invalid value: "{method}" for method used. '
            'The value must be "optgp" or "achr".'
        )

    return pd.DataFrame(
        columns=[rxn.id for rxn in model.reactions], data=sampler.sample(n)
    )



"""Provide functions for I/O in YAML format."""

import io
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from ruamel.yaml.compat import StringIO
from ruamel.yaml.main import YAML

from .dict import model_from_dict, model_to_dict


if TYPE_CHECKING:
    from io import TextIOBase

    from cobra import Model


YAML_SPEC = "1.2"


class CobraYAML(YAML):
    """Define custom subclass for YAML I/O."""

    def dump(
        self, data: Dict, stream: Optional["TextIOBase"] = None, **kwargs: Any
    ) -> str:
        """Dump YAML data.

        Parameters
        ----------
        data : dict
            A dictionary representing the cobra model and its components.
        stream : TextIOBase, optional
            A text stream object inheriting from `io.TextIOBase`. If None,
            `ruamel.yaml.compat.StringIO` is used (default None).
        **kwargs : Any
            Keyword arguments passed on to `ruamel.yaml.main.YAML.dump`.

        Returns
        -------
        str
            YAML string representation of `data`.

        """
        if stream is None:
            inefficient = True
            stream = StringIO()
        else:
            inefficient = False
        YAML.dump(self, data, stream, **kwargs)
        if inefficient:
            return stream.getvalue()


yaml = CobraYAML(typ="rt")


def to_yaml(model: "Model", sort: bool = False, **kwargs: Any) -> str:
    """Return the model as a YAML string.

    Parameters
    ----------
    model : cobra.Model
        The cobra model to represent.
    sort : bool, optional
        Whether to sort the metabolites, reactions, and genes or maintain the
        order defined in the model (default False).
    **kwargs : Any
        Keyword arguments passed on to `CobraYAML.dump`.

    Returns
    -------
    str
        YAML string representation of the cobra model.

    See Also
    --------
    save_yaml_model : Write directly to a file.
    ruamel.yaml.dump : Base function.

    """
    obj = model_to_dict(model, sort=sort)
    obj["version"] = YAML_SPEC
    return yaml.dump(obj, **kwargs)


def from_yaml(document: str) -> "Model":
    """Load a cobra model from a YAML string.

    Parameters
    ----------
    document : str
        The YAML string representation of a cobra model.

    Returns
    -------
    cobra.Model
        The cobra model as interpreted from the YAML document.

    See Also
    --------
    load_yaml_model : Load directly from a file.

    """
    content = StringIO(document)
    return model_from_dict(yaml.load(content))


def save_yaml_model(
    model: "Model", filename: str, sort: bool = False, **kwargs: Any
) -> None:
    """Write the cobra model to a file in YAML format.

    Parameters
    ----------
    model : cobra.Model
        The cobra model to represent.
    filename : str or file-like
        File path or descriptor that the YAML representation should be
        written to.
    sort : bool, optional
        Whether to sort the metabolites, reactions, and genes or maintain the
        order defined in the model (default False).
    **kwargs : Any
        Keyword arguments passed to `CobraYAML.dump`.

    See Also
    --------
    to_yaml : Return a string representation.
    ruamel.yaml.dump : Base function.

    """
    obj = model_to_dict(model, sort=sort)
    obj["version"] = YAML_SPEC
    if isinstance(filename, str):
        with io.open(filename, "w") as file_handle:
            yaml.dump(obj, file_handle, **kwargs)
    else:
        yaml.dump(obj, filename, **kwargs)


def load_yaml_model(filename: Union[str, Path]) -> "Model":
    """Load a cobra model from a file in YAML format.

    Parameters
    ----------
    filename : str or file-like
        File path or descriptor that contains the YAML document describing the
        cobra model.

    Returns
    -------
    cobra.Model
        The cobra model as represented in the YAML document.

    See Also
    --------
    from_yaml : Load from a string.

    """
    if isinstance(filename, (str, Path)):
        with io.open(filename, "r") as file_handle:
            return model_from_dict(yaml.load(file_handle))
    else:
        return model_from_dict(yaml.load(filename))



"""Provide functions for pruning reactions, metabolites and genes."""
import logging
from ast import And, BoolOp, Name, NodeTransformer
from functools import partial
from typing import TYPE_CHECKING, Iterable, List, Optional, Set, Tuple, Union

from cobra.util import get_context


if TYPE_CHECKING:
    from cobra import Gene, Metabolite, Model, Reaction

logger = logging.getLogger(__name__)


def prune_unused_metabolites(model: "Model") -> Tuple["Model", List["Metabolite"]]:
    """Remove metabolites not involved in any reactions.

    Parameters
    ----------
    model: cobra.Model
        The model to remove unused metabolites from.

    Returns
    -------
    cobra.Model
        Input model with unused metabolites removed.
    list of cobra.Metabolite
        List of metabolites that were removed.

    """
    output_model = model.copy()
    inactive_metabolites = [
        m for m in output_model.metabolites if len(m.reactions) == 0
    ]
    output_model.remove_metabolites(inactive_metabolites)
    return output_model, inactive_metabolites


def prune_unused_reactions(model: "Model") -> Tuple["Model", List["Reaction"]]:
    """Remove reactions with no assigned metabolites, returns pruned model.

    Parameters
    ----------
    model: cobra.Model
        The model to remove unused reactions from.

    Returns
    -------
    cobra.Model
        Input model with unused reactions removed.
    list of cobra.Reaction
        List of reactions that were removed.

    """
    output_model = model.copy()
    reactions_to_prune = [r for r in output_model.reactions if len(r.metabolites) == 0]
    output_model.remove_reactions(reactions_to_prune)
    return output_model, reactions_to_prune


def knock_out_model_genes(
    model: "Model",
    gene_list: Iterable[Union["Gene", int, str]],
) -> List["Reaction"]:
    """Disable the genes in `gene_list`.

    It sets the bounds to "zero" for reactions catalysed by the genes in
    `gene_list` if deleting the genes would stop the reactions from proceeding.

    The changes are reverted upon exit if executed within the model as context.

    Parameters
    ----------
    model: cobra.Model
        The model whose reaction bounds are to be set.
    gene_list: list of cobra.Gene
        The list of genes to knock-out.

    Returns
    -------
    list[cobra.Reaction]
        A list of cobra.Reactions that had the bounds turned to zero.
    """
    rxn_set = set()
    for gene in model.genes.get_by_any(gene_list):
        gene.knock_out()
        rxn_set.update(gene.reactions)
    return [rxn for rxn in rxn_set if not rxn.functional]


def delete_model_genes(
    model: "Model",
    gene_list: Union[List["Gene"], Set["Gene"], List[str], Set[str]],
    cumulative_deletions: bool = True,
    disable_orphans: bool = False,
) -> None:
    """Temporarily remove the effect of genes in `gene_list`.

    It sets the bounds to "zero" for reactions catalysed by the genes in
    `gene_list` if deleting the genes stops the reactions from proceeding.

    Parameters
    ----------
    model: cobra.Model
        The model whose reaction bounds are to be set.
    gene_list: list of cobra.Gene
        The list of genes to knock-out.
    cumulative_deletions: bool, optional
        If True, then any previous deletions will be maintained in the
        model (default True). Unused, ignored.
    disable_orphans: bool, optional
        If True, then orphan reactions will be disabled. Currently, this
        is not implemented (default False). Unused, ignored.

    .. deprecated :: 0.25
        Use cobra.manipulation.knock_out_model_genes to simulate knockouts
        and cobra.manipulation.remove_genes to remove genes from
        the model.

    See Also
    --------
    knock_out_model_genes()
    remove_model_genes()

    """
    logger.warning(
        "delete_model_genes and undelete_model_genes were deprecated "
        "and should not be used. undelete_model_genes was removed."
        "Use cobra.manipulation.remove_genes instead to remove genes "
        "from the model."
    )
    logger.warning(
        f"Use cobra.manipulation.knock_out_model_genes({model}, {gene_list})"
        f" to cause knockouts."
    )
    logger.warning(
        f"If you want to temporarily knock out genes (like undelete_model_genes) did, "
        f"run it in model context, for example"
        f"with({model}):"
        f"    knocked_out_rxns = knock_out_model_genes({model}, {gene_list})"
        f" since it will be reset when leaving the context."
    )


class _GeneRemover(NodeTransformer):
    """
    Class to represent a gene set remover.

    Parameters
    ----------
    target_genes: list or set of cobra.Gene
        A set of genes to be removed.

    """

    def __init__(self, target_genes: Set["Gene"], **kwargs) -> None:
        """Initialize a new object.

        Other Parameters
        ----------------
        kwargs:
            Further keyword arguments are passed on to the parent class.

        """
        super().__init__(**kwargs)
        self.target_genes = {str(i) for i in target_genes}

    def visit_Name(self, node: "Name") -> Optional["Name"]:
        """Remove a gene.

        Parameters
        ----------
        node: ast.Name
            The gene to remove.

        Returns
        -------
        cobra.Gene or None
            None if gene object is in `target_genes`.

        """
        return None if node.id in self.target_genes else node

    def visit_BoolOp(self, node: "BoolOp") -> Optional[Union["BoolOp", "Name"]]:
        """Rules for boolean operations.

        Parameters
        ----------
        node: ast.Name
            The gene to apply rules to.

        Returns
        -------
        ast.Name or None
            None if size of Or node values is zero after applying rule,
            or size of And node values is lower after applying rule.

        """
        original_n = len(node.values)
        self.generic_visit(node)
        if len(node.values) == 0:
            return None
        # AND with any entities removed
        if len(node.values) < original_n and isinstance(node.op, And):
            return None
        # if one entity in an OR was left, just that entity passed up
        if len(node.values) == 1:
            return node.values[0]
        return node


def remove_genes(
    model: "Model",
    gene_list: Union[List["Gene"], Set["Gene"], List[str], Union[str]],
    remove_reactions: bool = True,
) -> None:
    """Remove genes entirely from the model.

    This will also simplify all gene-reaction rules with the genes
    inactivated.

    Parameters
    ----------
    model: cobra.Model
        The model to remove genes from.
    gene_list: list of cobra.Gene or gene ids
        The list of gene objects to remove.
    remove_reactions: bool, optional
        Whether to remove reactions associated with genes in `gene_list`
        (default True).

    """
    gene_set = {model.genes.get_by_id(str(i)) for i in gene_list}
    gene_id_set = {i.id for i in gene_set}
    remover = _GeneRemover(gene_id_set)
    target_reactions = []
    rxns_to_revisit = set()
    context = get_context(model)
    for rxn in model.reactions:
        if rxn.gene_reaction_rule is None or len(rxn.gene_reaction_rule) == 0:
            continue
        # reactions to remove
        if remove_reactions and not rxn.gpr.eval(gene_id_set):
            target_reactions.append(rxn)
        else:
            # if the reaction is not removed, remove the gene
            # from its gpr
            old_gpr = rxn.gpr.copy()
            remover.visit(rxn.gpr)
            # If the remover completely removed the AST tree from the GPR, it will not
            # have body at all, which is why this isn't if body is None.
            if not hasattr(rxn.gpr, "body"):
                rxn.gpr.body = None
                rxn._genes = set()
            else:
                rxns_to_revisit.add(rxn)
            if context:
                context(partial(setattr, rxn, "gpr", old_gpr))
                context(partial(rxn.update_genes_from_gpr))
    for gene in gene_set:
        model.genes.remove(gene)
        gene._model = None
        if context:
            context(partial(model.genes.add, gene))
            context(partial(setattr, gene, "_model", model))
        # remove reference to the gene in all groups
        associated_groups = model.get_associated_groups(gene)
        for group in associated_groups:
            group.remove_members(gene)
    model.remove_reactions(target_reactions)
    for rxn in rxns_to_revisit:
        rxn.update_genes_from_gpr()



#!/usr/bin/env python


"""Set up the cobra package."""


from sys import version_info
from warnings import warn

from setuptools import setup


if version_info[:2] < (3, 6):
    warn(
        "We only explicitly test Python 3.6 and later, however, earlier versions may "
        "be functional if you can manage the dependencies yourself."
    )


# All other arguments are defined in `setup.cfg`.
if __name__ == "__main__":
    setup(version="0.26.3")



"""Define the group class."""

from typing import Iterable, Optional, Set
from warnings import warn

from .object import Object


class Group(Object):
    """
    Manage groups via this implementation of the SBML group specification.

    `Group` is a class for holding information regarding a pathways,
    subsystems, or other custom groupings of objects within a cobra.Model
    object.

    Parameters
    ----------
    id : str
        The identifier to associate with this group
    name : str, optional
        A human readable name for the group
    members : iterable, optional
        A DictList containing references to cobra.Model-associated objects
        that belong to the group.
    kind : {"collection", "classification", "partonomy"}, optional
        The kind of group, as specified for the Groups feature in the SBML
        level 3 package specification. Can be any of "classification",
        "partonomy", or "collection". The default is "collection".
        Please consult the SBML level 3 package specification to ensure you
        are using the proper value for kind. In short, members of a
        "classification" group should have an "is-a" relationship to the group
        (e.g. member is-a polar compound, or member is-a transporter).
        Members of a "partonomy" group should have a "part-of" relationship
        (e.g. member is part-of glycolysis). Members of a "collection" group
        do not have an implied relationship between the members, so use this
        value for kind when in doubt (e.g. member is a gap-filled reaction,
        or member is involved in a disease phenotype).
    """

    KIND_TYPES = ("collection", "classification", "partonomy")

    def __init__(
        self,
        id: str,
        name: str = "",
        members: Optional[Iterable] = None,
        kind: Optional[str] = None,
    ):
        """Initialize the group object.

         id : str
            The identifier to associate with this group
        name : str, optional
            A human readable name for the group
        members : iterable, optional
            A DictList containing references to cobra.Model-associated objects
            that belong to the group.
        kind : {"collection", "classification", "partonomy"}, optional
            The kind of group, as specified for the Groups feature in the SBML
            level 3 package specification.
        """
        Object.__init__(self, id, name)

        self._members = set() if members is None else set(members)
        self._kind = None
        self.kind = "collection" if kind is None else kind
        # self.model is None or refers to the cobra.Model that
        # contains self
        self._model = None

    def __len__(self) -> int:
        """Get length of group.

        Returns
        -------
        int
            An int with the length of the group.

        """
        return len(self._members)

    # read-only
    @property
    def members(self) -> Set:
        """Get members of the group.

        Returns
        -------
        Set
            A Set containing the members of the group.
        """
        return self._members

    @property
    def kind(self) -> str:
        """Return the group kind.

        Returns
        -------
        str
            The group kind. Should be one of the three types allowed in SBML.

        """
        return self._kind

    @kind.setter
    def kind(self, kind: str) -> None:
        """Set the group kind.

        Parameters
        ----------
        kind: str
            Must be one of the allowed kind types "collection", "classification",
            "partonomy".
            If kind is not one of these types, a ValueError Exception is raised.
        """
        kind = kind.lower()
        if kind in self.KIND_TYPES:
            self._kind = kind
        else:
            raise ValueError(f"Kind can only by one of: {', '.join(self.KIND_TYPES)}.")

    def add_members(self, new_members: list) -> None:
        """Add objects to the group.

        Parameters
        ----------
        new_members : list
            A list of cobrapy objects to add to the group.
        """
        if isinstance(new_members, str) or hasattr(new_members, "id"):
            warn("need to pass in a list")
            new_members = [new_members]

        self._members.update(new_members)

    def remove_members(self, to_remove: list) -> None:
        """Remove objects from the group.

        Parameters
        ----------
        to_remove : list
            A list of cobra objects to remove from the group
        """
        if isinstance(to_remove, str) or hasattr(to_remove, "id"):
            warn("need to pass in a list")
            to_remove = [to_remove]

        self._members.difference_update(to_remove)



"""Provide the ACHR sampler class."""

from typing import TYPE_CHECKING, Optional

import numpy as np
import pandas as pd

from .core import step
from .hr_sampler import HRSampler


if TYPE_CHECKING:
    from cobra import Model


class ACHRSampler(HRSampler):
    """
    Artificial Centering Hit-and-Run sampler.

    A sampler with low memory footprint and good convergence.

    Parameters
    ----------
    model : cobra.Model
        The cobra model from which to generate samples.
    thinning : int, optional
        The thinning factor of the generated sampling chain. A thinning of
        10 means samples are returned every 10 steps (default 100).
    nproj : int > 0, optional
        How often to reproject the sampling point into the feasibility
        space. Avoids numerical issues at the cost of lower sampling. If
        you observe many equality constraint violations with
        `sampler.validate` you should lower this number (default None).
    seed : int > 0, optional
        Sets the random number seed. Initialized to the current time stamp
        if None (default None).

    Attributes
    ----------
    n_samples : int
        The total number of samples that have been generated by this
        sampler instance.
    problem : typing.NamedTuple
        A NamedTuple whose attributes define the entire sampling problem in
        matrix form.
    warmup : numpy.matrix
        A numpy matrix with as many columns as reactions in the model and
        more than 3 rows containing a warmup sample in each row. None if no
        warmup points have been generated yet.
    retries : int
        The overall of sampling retries the sampler has observed. Larger
        values indicate numerical instabilities.
    fwd_idx : numpy.array
        A numpy array having one entry for each reaction in the model,
        containing the index of the respective forward variable.
    rev_idx : numpy.array
        A numpy array having one entry for each reaction in the model,
        containing the index of the respective reverse variable.
    prev : numpy.array
        The current/last flux sample generated.
    center : numpy.array
        The center of the sampling space as estimated by the mean of all
        previously generated samples.

    Notes
    -----
    ACHR generates samples by choosing new directions from the sampling
    space's center and the warmup points. The implementation used here is
    the same as in the MATLAB COBRA Toolbox [2]_ and uses only the initial
    warmup points to generate new directions and not any other previous
    iterations. This usually gives better mixing, since the startup points
    are chosen to span the space in a wide manner. This also makes the
    generated sampling chain quasi-Markovian since the center converges
    rapidly.

    Memory usage is roughly in the order of (2 * number of reactions) ^ 2
    due to the required nullspace matrices and warmup points. So, large
    models easily take up a few GBs of RAM.

    References
    ----------
    .. [1] Direction Choice for Accelerated Convergence in Hit-and-Run Sampling
       David E. Kaufman, Robert L. Smith
       Operations Research 199846:1 , 84-95
       https://doi.org/10.1287/opre.46.1.84

    .. [2] https://github.com/opencobra/cobratoolbox

    """

    def __init__(
        self,
        model: "Model",
        thinning: int = 100,
        nproj: Optional[int] = None,
        seed: Optional[int] = None,
        **kwargs
    ) -> None:
        """Initialize a new ACHRSampler."""
        super().__init__(model, thinning, nproj=nproj, seed=seed, **kwargs)
        self.generate_fva_warmup()
        self.prev = self.center = self.warmup.mean(axis=0)
        np.random.seed(self._seed)

    def __single_iteration(self) -> None:
        """Run a single iteration of the sampling."""
        pi = np.random.randint(self.n_warmup)
        # mix in the original warmup points to not get stuck
        delta = self.warmup[pi, :] - self.center
        self.prev = step(self, self.prev, delta)

        if self.problem.homogeneous and (
            self.n_samples * self.thinning % self.nproj == 0
        ):
            self.prev = self._reproject(self.prev)
            self.center = self._reproject(self.center)

        self.center = (self.n_samples * self.center) / (
            self.n_samples + 1
        ) + self.prev / (self.n_samples + 1)
        self.n_samples += 1

    def sample(self, n: int, fluxes: bool = True) -> pd.DataFrame:
        """Generate a set of samples.

        This is the basic sampling function for all hit-and-run samplers.

        Parameters
        ----------
        n : int
            The number of samples that are generated at once.
        fluxes : bool, optional
            Whether to return fluxes or the internal solver variables. If
            set to False, will return a variable for each forward and
            backward flux as well as all additional variables you might
            have defined in the model (default True).

        Returns
        -------
        pandas.DataFrame
            Returns a pandas DataFrame with `n` rows, each containing a
            flux sample.

        Notes
        -----
        Performance of this function linearly depends on the number
        of reactions in your model and the thinning factor.

        """
        samples = np.zeros((n, self.warmup.shape[1]))

        for i in range(1, self.thinning * n + 1):
            self.__single_iteration()

            if i % self.thinning == 0:
                samples[i // self.thinning - 1, :] = self.prev

        if fluxes:
            names = [r.id for r in self.model.reactions]

            return pd.DataFrame(
                samples[:, self.fwd_idx] - samples[:, self.rev_idx],
                columns=names,
            )
        else:
            names = [v.name for v in self.model.variables]

            return pd.DataFrame(samples, columns=names)
